import bz2
import gzip
from six.moves import builtins
from unittest import TestCase
from unittest.mock import mock_open, patch
from six import assertRaisesRegex
from collections import Counter
from io import BytesIO, StringIO

from dark.utils import (
    numericallySortFilenames,
    median,
    asHandle,
    parseRangeString,
    parseRangeExpression,
    pct,
    baseCountsToStr,
    nucleotidesToStr,
    countPrint,
    take,
    matchOffset,
)


class TestNumericallySortFilenames(TestCase):
    """
    Test the numericallySortFilenames function.
    """

    def testNoNames(self):
        """
        An empty list must be returned when an empty list is given.
        """
        self.assertEqual([], numericallySortFilenames([]))

    def testOneNonNumericName(self):
        """
        A list with a single non-numeric name should result in that same
        name being returned.
        """
        self.assertEqual(["hey"], numericallySortFilenames(["hey"]))

    def testOneNumericName(self):
        """
        A list with a single numeric name should result in that same
        name being returned.
        """
        self.assertEqual(["3.json"], numericallySortFilenames(["3.json"]))

    def testSeveralNames(self):
        """
        A list with several numeric names should result in a correctly
        sorted list of names being returned.
        """
        self.assertEqual(
            ["1.json", "2.json", "3.json"],
            numericallySortFilenames(["3.json", "1.json", "2.json"]),
        )

    def testSeveralNamesWithUnequalPrefixLengths(self):
        """
        A list with several numeric names whose numeric prefixes differ
        in length should result in a correctly sorted list of names being
        returned.
        """
        self.assertEqual(
            ["2.json", "3.json", "21.json", "35.json", "250.json"],
            numericallySortFilenames(
                ["3.json", "21.json", "35.json", "250.json", "2.json"]
            ),
        )

    def testBasename(self):
        """
        Sorting must be according to file basename.
        """
        self.assertEqual(
            [
                "../output/2.json",
                "../output/3.json",
                "../output/21.json",
                "../output/35.json",
                "../output/250.json",
            ],
            numericallySortFilenames(
                [
                    "../output/3.json",
                    "../output/21.json",
                    "../output/35.json",
                    "../output/250.json",
                    "../output/2.json",
                ]
            ),
        )


class TestMedian(TestCase):
    """
    Tests for the median function.
    """

    def testEmptyArgRaises(self):
        """
        An empty list must cause median to raise ValueError.
        """
        error = "^arg is an empty sequence$"
        assertRaisesRegex(self, ValueError, error, median, [])

    def testMedianOfOne(self):
        """
        The median function must work on a list of length one.
        """
        self.assertEqual(3, median([3]))

    def testMedianOfTwo(self):
        """
        The median function must work on a list of length two.
        """
        self.assertEqual(4.5, median([3.1, 5.9]))

    def testMedianOfThree(self):
        """
        The median function must work on a list of length threee.
        """
        self.assertEqual(5.9, median([3.1, 7.6, 5.9]))

    def testMedianOfFour(self):
        """
        The median function must work on a list of length four.
        """
        self.assertEqual(4.5, median([3.1, 1.3, 7.6, 5.9]))

    def testMedianOfFive(self):
        """
        The median function must work on a list of length five.
        """
        self.assertEqual(5.9, median([3.1, 1.3, 7.6, 9.9, 5.9]))


class TestAsHandle(TestCase):
    """
    Test the asHandle function
    """

    def testOpenFile(self):
        """
        When an open file pointer is passed to asHandle, that same file
        pointer must be returned.
        """
        with patch.object(builtins, "open", mock_open()):
            fp = open("file")
            with asHandle(fp) as newfp:
                self.assertIs(fp, newfp)

    def testStr(self):
        """
        When a string filename is passed to asHandle, it must be possible to
        read the correct data from the fp that is returned.
        """
        mockOpener = mock_open(read_data="xxx")
        with patch.object(builtins, "open", mockOpener):
            with asHandle("file") as fp:
                self.assertEqual("xxx", fp.read())

    def testBZ2(self):
        """
        When a string '*.bz2' filename is passed to asHandle, it must be
        possible to read the correct data from the fp that is returned.
        """
        result = BytesIO(b"xxx")

        with patch.object(bz2, "BZ2File") as mockMethod:
            mockMethod.return_value = result
            with asHandle("file.bz2") as fp:
                self.assertEqual("xxx", fp.read())

    def testGzip(self):
        """
        When a string '*.gz' filename is passed to asHandle, it must be
        possible to read the correct data from the fp that is returned.
        """
        result = BytesIO(b"xxx")

        with patch.object(gzip, "GzipFile") as mockMethod:
            mockMethod.return_value = result
            with asHandle("file.gz") as fp:
                self.assertEqual("xxx", fp.read())


class TestParseRangeString(TestCase):
    """
    Check that the parseRangeString function works as expected.
    """

    def testEmptyString(self):
        """
        An empty string must produce an empty set of indices.
        """
        error = "^Illegal range ''. Ranges must single numbers or " "number-number\\.$"
        assertRaisesRegex(self, ValueError, error, parseRangeString, "")

    def testSingleNumber(self):
        """
        A single number must result in the expected set.
        """
        self.assertEqual({6}, parseRangeString("6"))

    def testSingleNumberSpaceBefore(self):
        """
        A single number preceeded by whitespace must result in the expected
        set.
        """
        self.assertEqual({6}, parseRangeString("  6"))

    def testSingleNumberSpaceAfter(self):
        """
        A single number followed by whitespace must result in the expected
        set.
        """
        self.assertEqual({6}, parseRangeString("6  "))

    def testSingleNumberSpaceBeforeAndAfter(self):
        """
        A single number preceeded and followed by whitespace must result in
        the expected set.
        """
        self.assertEqual({6}, parseRangeString(" 6  "))

    def testSingleRange(self):
        """
        A single range must result in the expected set.
        """
        self.assertEqual({6, 7, 8, 9, 10}, parseRangeString("6-10"))

    def testSingleRangeWithSpaceBeforeHyphen(self):
        """
        A single range with a space before the hyphen must result in the
        expected set.
        """
        self.assertEqual({6, 7, 8, 9, 10}, parseRangeString("6 -10"))

    def testSingleRangeWithSpaceAfterHyphen(self):
        """
        A single range with a space after the hyphen must result in the
        expected set.
        """
        self.assertEqual({6, 7, 8, 9, 10}, parseRangeString("6- 10"))

    def testSingleRangeWithSpaceBeforeAfterHyphen(self):
        """
        A single range with spaces before and after the hyphen must result in
        the expected set.
        """
        self.assertEqual({6, 7, 8, 9, 10}, parseRangeString("6 - 10"))

    def testTwoRanges(self):
        """
        Two ranges must result in the expected set.
        """
        self.assertEqual({6, 7, 8, 9, 10}, parseRangeString("6-8,9-10"))

    def testTwoOverlappingRanges(self):
        """
        Two overlapping ranges must result in the expected set.
        """
        self.assertEqual({6, 7, 8, 9, 10}, parseRangeString("6-9,7-10"))

    def testTwoRangesAndANumber(self):
        """
        Two ranges and a number must result in the expected set.
        """
        self.assertEqual({6, 7, 8, 10}, parseRangeString("6-8,10"))

    def testTwoRangesAndTwoNumbers(self):
        """
        Two ranges and two numbers must result in the expected set.
        """
        self.assertEqual({4, 6, 7, 8, 9, 10, 11, 12}, parseRangeString("6-8,9,10-12,4"))

    def testZeroConversion(self):
        """
        If we ask for zero conversion, the result must be as expected.
        """
        self.assertEqual(
            {3, 5, 6, 7, 8, 9, 10, 11},
            parseRangeString("6-8,9,10-12,4", convertToZeroBased=True),
        )


class TestParseRangeExpression(TestCase):
    """
    Check that the parseRangeExpression function works as expected.
    """

    def testInvalidExpression(self):
        """
        An invalid string must raise a ValueError.
        """
        error = r"^\($"
        assertRaisesRegex(self, ValueError, error, parseRangeExpression, "(")
        error = r"^hey$"
        assertRaisesRegex(self, ValueError, error, parseRangeExpression, "hey")

    def testEmptyString(self):
        """
        An empty string must produce an empty set.
        """
        self.assertEqual(set(), parseRangeExpression(""))

    def testOneRange(self):
        """
        A simple 3-4 string must produce the expected set.
        """
        self.assertEqual({3, 4}, parseRangeExpression("3-4"))

    def testOneRangeZeroBased(self):
        """
        A simple 3-4 string must produce the expected set when
        convertToZeroBased is True.
        """
        self.assertEqual({2, 3}, parseRangeExpression("3-4", True))

    def testCommas(self):
        """
        A simple 3,4,5 string must produce the expected set.
        """
        self.assertEqual({3, 4, 5}, parseRangeExpression("3,4,5"))

    def testCommasAndRange(self):
        """
        A simple 3,4,5-7 string must produce the expected set.
        """
        self.assertEqual({3, 4, 5, 6, 7}, parseRangeExpression("3,4,5-7"))

    def testTwoRanges(self):
        """
        A simple 3-4,6-8 string must produce the expected set.
        """
        self.assertEqual({3, 4, 6, 7, 8}, parseRangeExpression("3-4,6-8"))

    def testTwoRangesWithSpace(self):
        """
        A simple 3-4, 6-8 string must produce the expected set.
        """
        self.assertEqual({3, 4, 6, 7, 8}, parseRangeExpression("3-4, 6-8"))

    def testUnion(self):
        """
        A union such as 3-4 | 6-8 must produce the expected set.
        """
        self.assertEqual({3, 4, 6, 7, 8}, parseRangeExpression("3-4 | 6-8"))

    def testIntersection(self):
        """
        An intersection such as 3-4 & 4-8 must produce the expected set.
        """
        self.assertEqual({4}, parseRangeExpression("3-4 & 4-8"))

    def testDifferenceNoSpaces(self):
        """
        A difference such as 6-10-7-8 must produce the expected set.
        """
        self.assertEqual({6, 9, 10}, parseRangeExpression("6-10-7-8"))

    def testDifferenceWithSpaces(self):
        """
        A difference such as 6-10 - 7-8 must produce the expected set.
        """
        self.assertEqual({6, 9, 10}, parseRangeExpression("6-10 - 7-8"))

    def testParens(self):
        """
        A difference with parentheses such as '(3-5 | 7-9) & 5-7' must produce
        the expected set.
        """
        self.assertEqual({5, 7}, parseRangeExpression("(3-5 | 7-9) & 5-7"))

    def testDoubleParens(self):
        """
        A difference with two parentheses such as '(3-5 | 7-9) & (5-7 | 9-11)'
        must produce the expected set.
        """
        self.assertEqual({5, 7, 9}, parseRangeExpression("(3-5 | 7-9) & (5-7 | 9-11)"))


class TestStringIO(TestCase):
    """
    Tests for our StringIO class.
    """

    def testInitiallyEmpty(self):
        """
        A StringIO instance must initially be empty.
        """
        self.assertEqual("", StringIO().getvalue())

    def testWriteRead(self):
        """
        It must be possible to write and read to/from a StringIO instance as
        normal.
        """
        s = StringIO()
        s.write("hey")
        self.assertEqual("hey", s.getvalue())

    def testInitializedRead(self):
        """
        It must be possible to read from a StringIO instance that is
        initialized on creation.
        """
        s = StringIO("hey")
        self.assertEqual("hey", s.getvalue())

    def testContextManager(self):
        """
        It must be possible to use a StringIO instance as a context manager.
        """
        with StringIO() as s:
            s.write("hey")
            self.assertEqual("hey", s.getvalue())


class TestBaseCountsToStr(TestCase):
    """
    Test the baseCountsToStr function.
    """

    def testSimple(self):
        """
        A simple example must work as expected.
        """
        counts = Counter()
        counts["A"] += 1
        counts["G"] += 2
        self.assertEqual("A:1 G:2", baseCountsToStr(counts))


class TestNucleotidesToStr(TestCase):
    """
    Test the nucleotidesToStr function.
    """

    def testSimple(self):
        """
        A simple example must work as expected.
        """
        counts1 = Counter()
        counts1["A"] += 1
        counts1["G"] += 2
        counts2 = Counter()
        counts2["C"] += 1
        counts2["T"] += 3
        self.assertEqual(
            "0: A:1 G:2\n7: C:1 T:3",
            nucleotidesToStr(
                {
                    0: counts1,
                    7: counts2,
                }
            ),
        )


class TestCountPrint(TestCase):
    """
    Test the countPrint function and the contained percentage function.
    """

    def testSimple(self):
        """
        A simple example must work as expected.
        """
        count = 2
        len1 = 10
        self.assertEqual("Count is: 2/10 (20.00%)", countPrint("Count is", count, len1))

    def testTwoSequences(self):
        """
        An example involving two different lengths must work as expected.
        """
        count = 2
        len1 = 10
        len2 = 8
        self.assertEqual(
            "Count is: 2/10 (20.00%) of sequence 1," " 2/8 (25.00%) of sequence 2",
            countPrint("Count is", count, len1, len2),
        )


class TestPct(TestCase):
    """
    Test the pct function.
    """

    def testZeroNumerator(self):
        """
        The pct function must produce the correct result if the numerator is
        zero.
        """
        self.assertEqual("0/10 (0.000%)", pct(0, 10))

    def testZeroDenominator(self):
        """
        The pct function must produce the correct result if the denominator is
        zero.
        """
        self.assertEqual("0/0 (0.000%)", pct(0, 0))

    def testOneHalf(self):
        """
        The pct function must produce the correct result if the numerator is
        one half of the denominator.
        """
        self.assertEqual("5/10 (50.000%)", pct(5, 10))

    def testOneSeventh(self):
        """
        The pct function must produce the correct result if the numerator is
        one seventh of the denominator.
        """
        self.assertEqual("2/14 (14.286%)", pct(2, 14))


class TestTake(TestCase):
    """
    Test the take function.
    """

    def testNLessThanOne(self):
        """
        The take function must raise an AssertionError if passed n < 1.
        list.
        """
        self.assertRaises(AssertionError, next, take([], 0))

    def testEmpty(self):
        """
        The take function must return an empty generator if passed an empty
        list.
        """
        self.assertEqual([], list(take([], 4)))

    def testOne(self):
        """
        The take function must return individual items if passed n=1.
        """
        self.assertEqual([[3], [4], [5]], list(take([3, 4, 5], 1)))

    def testNBiggerThanPassedList(self):
        """
        The take function must return all items if passed an n that is bigger
        than the passed list.
        """
        self.assertEqual([[3, 4, 5]], list(take([3, 4, 5], 100)))

    def testThree(self):
        """
        The take function must return three items at a time if passed n=3, and
        also return the extra two items.
        """
        self.assertEqual(
            [[0, 1, 2], [3, 4, 5], [6, 7, 8], [9, 10]], list(take(range(11), 3))
        )


class TestMatchOffset(TestCase):
    """
    Test the matchOffset function.
    """

    def testEmpty(self):
        """
        An empty query must match an empty reference at position 0.
        """
        self.assertEqual(0, matchOffset("", ""))

    def testEqualStrings(self):
        """
        An non-empty reference must match an identical non-empty query at
        position 0.
        """
        self.assertEqual(0, matchOffset("AA", "AA"))

    def testQueryPaddedLeftByOne(self):
        """
        A query that is padded on the left by one space must match at
        position 1.
        """
        self.assertEqual(1, matchOffset("AA", " A"))

    def testQueryPaddedLeftByTwo(self):
        """
        A query that is padded on the left by two spaces must match at
        position 2.
        """
        self.assertEqual(2, matchOffset("AAA", "  A"))

    def testQueryPaddedLeftByTwoReferencePaddedLeftByOne(self):
        """
        A query that is padded on the left by two spaces must match a reference
        that is padded on the left by one space at position 1.
        """
        self.assertEqual(1, matchOffset(" AA", "  A"))

    def testQueryPaddedLeftByFiveReferencePaddedLeftByOneWithGaps(self):
        """
        A query that is padded on the left by five spaces must correctly match
        a reference containing gaps that is padded on the left by one space.
        """
        self.assertEqual(2, matchOffset(" AA--G", "     A"))

    def testQueryPaddedLeftByFiveReferenceOneGapLeftWithGaps(self):
        """
        A query that is padded on the left by five spaces must correctly match
        a reference that starts with a gap and that contains gaps.
        """
        self.assertEqual(2, matchOffset("-AA--G", "     A"))

    def testOmicronPartialInsertionRead(self):
        """
        A query that overlaps part of an insertion in the reference must be
        handled correctly.
        """
        self.assertEqual(
            12,
            matchOffset(
                "TAATTTAGTGCG---------TGATCTCCCTCAGGGTTTTTCGGCTTTAGAAC",
                "             AGCCAGAATGATCTCCCTCAGGGTTTTTCGGCTTT",
            ),
        )



from dark.blast.hsp import printHSP


def printBlastRecord(record):
    """
    Print a BLAST record.

    @param record: A BioPython C{Bio.Blast.Record.Blast} instance.
    """
    for key in sorted(record.__dict__):
        if key not in ["alignments", "descriptions", "reference"]:
            print("%s: %r" % (key, record.__dict__[key]))
    print("alignments: (%d in total):" % len(record.alignments))
    for i, alignment in enumerate(record.alignments):
        print("  description %d:" % (i + 1))
        for attr in ["accession", "bits", "e", "num_alignments", "score"]:
            print("    %s: %s" % (attr, getattr(record.descriptions[i], attr)))
        print("  alignment %d:" % (i + 1))
        for attr in "accession", "hit_def", "hit_id", "length", "title":
            print("    %s: %s" % (attr, getattr(alignment, attr)))
        print("    HSPs (%d in total):" % len(alignment.hsps))
        for hspIndex, hsp in enumerate(alignment.hsps, start=1):
            print("      hsp %d:" % hspIndex)
            printHSP(hsp, "        ")



from unittest import TestCase

from dark.hsp import HSP, LSP


class TestHSP(TestCase):
    """
    Tests of the L{dark.hsp.HSP} class.
    """

    def testExpectedAttributes(self):
        """
        An HSP must have the expected attributes.
        """
        hsp = HSP(
            7,
            readStart=1,
            readEnd=2,
            readStartInSubject=3,
            readEndInSubject=4,
            subjectStart=5,
            subjectEnd=6,
            readMatchedSequence="aaa",
            subjectMatchedSequence="ccc",
            readFrame=8,
            subjectFrame=9,
        )
        self.assertEqual(1, hsp.readStart)
        self.assertEqual(2, hsp.readEnd)
        self.assertEqual(3, hsp.readStartInSubject)
        self.assertEqual(4, hsp.readEndInSubject)
        self.assertEqual(5, hsp.subjectStart)
        self.assertEqual(6, hsp.subjectEnd)
        self.assertEqual("aaa", hsp.readMatchedSequence)
        self.assertEqual("ccc", hsp.subjectMatchedSequence)
        self.assertEqual(7, hsp.score.score)
        self.assertEqual(8, hsp.readFrame)
        self.assertEqual(9, hsp.subjectFrame)

    def testEqual(self):
        """
        Two HSPs must compare properly with ==
        """
        self.assertEqual(HSP(7), HSP(7))

    def testLt(self):
        """
        Two HSPs must compare properly with <
        """
        self.assertTrue(HSP(7) < HSP(8))

    def testBetterThanTrue(self):
        """
        Two HSP scores must compare properly with betterThan if the passed
        score is worse than the score of the HSP.
        """
        self.assertTrue(HSP(7).betterThan(5))

    def testBetterThanFalse(self):
        """
        Two HSP scores must compare properly with betterThan if the passed
        score is better than the score of the HSP.
        """
        self.assertFalse(HSP(5).betterThan(7))

    def testToDict(self):
        """
        The toDict method must return the expected dictionary.
        """
        hsp = HSP(
            0,
            readStart=1,
            readEnd=2,
            readStartInSubject=3,
            readEndInSubject=4,
            subjectStart=5,
            subjectEnd=6,
            readMatchedSequence="aaa",
            subjectMatchedSequence="ccc",
            readFrame=7,
            subjectFrame=8,
            identicalCount=9,
            percentIdentical=99.3,
            positiveCount=10,
            percentPositive=3.0,
        )

        self.assertEqual(
            {
                "score": 0,
                "readStart": 1,
                "readEnd": 2,
                "readStartInSubject": 3,
                "readEndInSubject": 4,
                "subjectStart": 5,
                "subjectEnd": 6,
                "readFrame": 7,
                "subjectFrame": 8,
                "identicalCount": 9,
                "percentIdentical": 99.3,
                "positiveCount": 10,
                "percentPositive": 3.0,
                "readMatchedSequence": "aaa",
                "subjectMatchedSequence": "ccc",
            },
            hsp.toDict(),
        )


class TestLSP(TestCase):
    """
    Tests of the L{dark.hsp.blast.LSP} class.
    """

    def testExpectedAttributes(self):
        """
        An LSP must have the expected attributes.
        """
        lsp = LSP(
            7,
            readStart=1,
            readEnd=2,
            readStartInSubject=3,
            readEndInSubject=4,
            subjectStart=5,
            subjectEnd=6,
            readMatchedSequence="aaa",
            subjectMatchedSequence="ccc",
            readFrame=8,
            subjectFrame=9,
        )
        self.assertEqual(1, lsp.readStart)
        self.assertEqual(2, lsp.readEnd)
        self.assertEqual(3, lsp.readStartInSubject)
        self.assertEqual(4, lsp.readEndInSubject)
        self.assertEqual(5, lsp.subjectStart)
        self.assertEqual(6, lsp.subjectEnd)
        self.assertEqual("aaa", lsp.readMatchedSequence)
        self.assertEqual("ccc", lsp.subjectMatchedSequence)
        self.assertEqual(7, lsp.score.score)
        self.assertEqual(8, lsp.readFrame)
        self.assertEqual(9, lsp.subjectFrame)

    def testEqual(self):
        """
        Two LSPs must compare properly with ==
        """
        self.assertEqual(LSP(7), LSP(7))

    def testLt(self):
        """
        Two LSPs must compare properly with <
        """
        self.assertTrue(LSP(8) < LSP(7))

    def testBetterThanTrue(self):
        """
        Two LSP scores must compare properly with betterThan if the passed
        score is worse than the score of the LSP.
        """
        self.assertTrue(LSP(5).betterThan(7))

    def testBetterThanFalse(self):
        """
        Two LSP scores must compare properly with betterThan if the passed
        score is better than the score of the LSP.
        """
        self.assertFalse(LSP(7).betterThan(5))

    def testToDict(self):
        """
        The toDict method must return the expected dictionary.
        """
        lsp = LSP(
            0,
            readStart=1,
            readEnd=2,
            readStartInSubject=3,
            readEndInSubject=4,
            subjectStart=5,
            subjectEnd=6,
            readMatchedSequence="aaa",
            subjectMatchedSequence="ccc",
            readFrame=7,
            subjectFrame=8,
            identicalCount=9,
            percentIdentical=99.3,
            positiveCount=10,
            percentPositive=9.9,
        )

        self.assertEqual(
            {
                "score": 0,
                "readStart": 1,
                "readEnd": 2,
                "readStartInSubject": 3,
                "readEndInSubject": 4,
                "subjectStart": 5,
                "subjectEnd": 6,
                "readFrame": 7,
                "subjectFrame": 8,
                "identicalCount": 9,
                "percentIdentical": 99.3,
                "positiveCount": 10,
                "percentPositive": 9.9,
                "readMatchedSequence": "aaa",
                "subjectMatchedSequence": "ccc",
            },
            lsp.toDict(),
        )



import sys
from os.path import join
from subprocess import CalledProcessError
from tempfile import mkdtemp
from shutil import rmtree
from typing import List, Optional, Tuple

import edlib

from dark.dna import AMBIGUOUS
from dark.fasta import FastaReads
from dark.process import Executor
from dark.reads import Read, Reads


EDLIB_AMBIGUOUS: Tuple[Tuple[str, ...], ...] = tuple(
    [
        tuple(sorted((nt, code)))
        for nt, ambiguities in AMBIGUOUS.items()
        for code in ambiguities
        if len(ambiguities) > 1
    ]
)


def mafft(
    reads: Reads,
    verbose: bool = False,
    options: Optional[str] = None,
    threads: Optional[int] = None,
    executor: Optional[Executor] = None,
    dryRun: bool = False,
) -> Reads:
    """
    Run a MAFFT alignment and return the sequences.

    @param reads: An iterable of multiple reads.
    @param verbose: If C{True} print progress info to sys.stderr.
    @param options: A C{str} of options to pass to mafft.
    @param executor: An C{Executor} instance. If C{None}, a new executor
        will be created and used.
    @param dryRun: If C{True}, do not execute commands, just let the executor
        log them.
    @return: A C{Reads} instance with the aligned sequences.
    """
    tempdir = mkdtemp()

    infile = join(tempdir, "input.fasta")
    out = join(tempdir, "result.fasta")

    Reads(reads).save(infile)

    if verbose:
        print("Running mafft.", file=sys.stderr)

    executor = executor or Executor(dryRun=dryRun)

    executor.execute(
        "mafft %s %s '%s' > '%s'"
        % (
            ("" if threads is None else "--thread %d" % threads),
            options or "",
            infile,
            out,
        ),
        dryRun=dryRun,
    )

    # Use 'list' in the following to force reading the FASTA from disk.
    result = Reads([] if dryRun else list(FastaReads(out)))
    rmtree(tempdir)

    return result


def needle(
    reads: List[Read],
    verbose: bool = False,
    options: Optional[str] = None,
    executor: Optional[Executor] = None,
    dryRun: bool = False,
) -> Reads:
    """
    Run a Needleman-Wunsch alignment and return the two sequences.

    @param reads: An iterable of two reads.
    @param verbose: If C{True} print progress info to sys.stderr.
    @param options: Additional options to pass to needle.
    @param executor: An C{Executor} instance. If C{None}, a new executor
        will be created and used.
    @param dryRun: If C{True}, do not execute commands, just let the executor
        log them.
    @return: A C{Reads} instance with the two aligned sequences.
    """
    tempdir = mkdtemp()

    file1 = join(tempdir, "file1.fasta")
    with open(file1, "w") as fp:
        print(reads[0].toString("fasta"), end="", file=fp)

    file2 = join(tempdir, "file2.fasta")
    with open(file2, "w") as fp:
        print(reads[1].toString("fasta"), end="", file=fp)

    out = join(tempdir, "result.fasta")

    def useStderr(e):
        return "Sequences too big. Try 'stretcher'" not in e.stderr

    executor = executor or Executor(dryRun=dryRun)

    if verbose:
        print("Running needle.", file=sys.stderr)
    try:
        executor.execute(
            "needle -asequence '%s' -bsequence '%s' %s "
            "-outfile '%s' -aformat fasta" % (file1, file2, options or "", out),
            dryRun=dryRun,
            useStderr=useStderr,
        )
    except CalledProcessError as e:
        if useStderr(e):
            raise
        else:
            if verbose:
                print(
                    "Sequences too long for needle. Falling back to "
                    "stretcher. Be patient!",
                    file=sys.stderr,
                )
            executor.execute(
                "stretcher -asequence '%s' -bsequence '%s' "
                "-auto "
                "-outfile '%s' -aformat fasta" % (file1, file2, out),
                dryRun=dryRun,
            )

    # Use 'list' in the following to force reading the FASTA from disk.
    result = Reads([] if dryRun else list(FastaReads(out)))
    rmtree(tempdir)

    return result


def removeFirstUnnecessaryGaps(
    seq1: str, seq2: str, gapSymbol: str = "-"
) -> Tuple[bool, str, str]:
    """
    Find and remove the first set of gaps in two sequences that can be removed
    without increasing the difference between the strings.

    @param seq1: A C{str} sequence string.
    @param seq2: A C{str} sequence string.
    @param gapSymbol: A C{str} 1-character symbol to use for gaps.
    @return: A 3-C{tuple} with either C{True} and two new C{str} sequences or
        C{False} and two empty strings if no removable gaps were found. The
        C{True} and C{False} first elements are just there to keep mypy happy,
        seeing as I couldn't figure out how to return (None, None) without
        getting warnings.
    """
    assert len(seq1) == len(seq2)
    for start in range(len(seq1)):
        if seq1[start] == gapSymbol:
            if seq2[start] == gapSymbol:
                raise ValueError(
                    f"Sequences {seq1!r} and {seq2!r} both have "
                    f"a gap ({gapSymbol!r}) at offset {start}!"
                )
            first, second = seq1, seq2
            break
        elif seq2[start] == gapSymbol:
            assert seq1[start] != gapSymbol
            first, second = seq2, seq1
            break
    else:
        # Did not find any gaps.
        return False, "", ""

    excessGapCount = 1
    for end in range(start + 1, len(first)):
        if first[end] == gapSymbol:
            assert second[end] != gapSymbol
            excessGapCount += 1
        elif second[end] == gapSymbol:
            excessGapCount -= 1
            if excessGapCount == 0:
                break
    else:
        # We did not get down to zero gaps. This will never happen if we
        # were originally called as a result of a call to edlibAlign
        # (below) and it only calls removeUnnecessaryGaps (also below) when
        # both original sequences have a different length from their
        # aligned versions. In that case there must be at least one gap in
        # both sequences.  But for now this is not considered an error
        # because we may not have been called by removeUnnecessaryGaps.
        return False, "", ""

    subseq1 = seq1[start : end + 1]
    subseq1noGaps = subseq1.replace(gapSymbol, "")
    subseq2 = seq2[start : end + 1]
    subseq2noGaps = subseq2.replace(gapSymbol, "")

    diffsWithGaps = sum(a != b for (a, b) in zip(subseq1, subseq2))
    diffsWithoutGaps = sum(a != b for (a, b) in zip(subseq1noGaps, subseq2noGaps))

    if diffsWithoutGaps <= diffsWithGaps:
        # The subsequences match at least as well without gaps, so replace
        # the gapped region in each sequence with its ungapped version.
        return (
            True,
            seq1[:start] + subseq1noGaps + seq1[end + 1 :],
            seq2[:start] + subseq2noGaps + seq2[end + 1 :],
        )
    else:
        return False, "", ""


def removeUnnecessaryGaps(
    seq1: str, seq2: str, gapSymbol: str = "-"
) -> Tuple[str, str]:
    """
    Find and remove all local sets of gaps in two sequences that can be removed
    without increasing the difference between the strings.

    @param seq1: A C{str} sequence string.
    @param seq2: A C{str} sequence string.
    @param gapSymbol: A C{str} 1-character symbol to use for gaps.
    @return: A 2-C{tuple} of C{str} sequences with removable gaps eliminated.
    """
    while True:
        foundGaps, new1, new2 = removeFirstUnnecessaryGaps(seq1, seq2, gapSymbol)
        if foundGaps:
            seq1, seq2 = new1, new2
        else:
            return seq1, seq2

    raise RuntimeError(
        "Fell off the end of removeFirstUnnecessaryGaps. " "This should be impossible!"
    )


def edlibAlign(
    reads: Reads,
    gapSymbol: str = "-",
    minimizeGaps: bool = True,
    onlyTwoSequences: bool = True,
    matchAmbiguous: bool = True,
) -> Reads:
    """
    Run an edlib alignment and return the sequences.

    @param reads: An iterable of at least two reads. The reads must not contain
        C{gapSymbol}.
    @param gapSymbol: A C{str} 1-character symbol to use for gaps.
    @param minimizeGaps: If C{True}, post-process the edlib output to remove
        unnecessary gaps.
    @param onlyTwoSequences: Ensure that C{reads} only has two reads.
    @param matchAmbiguous: If C{True}, count ambiguous nucleotides that are
        possibly correct as actually being correct. Otherwise, we are strict
        and nucleotide codes only match themselves.
    @raise ValueError: If C{onlyTwoSequences} is C{True} and there are more
       than two reads passed or if the gap symbol is already present in either
       of the reads.
    @return: A C{Reads} instance with the aligned sequences.
    """
    # Align the first two sequences.
    r1, r2, *rest = list(reads)

    # Complain if there were more than two sequences and we were told to be
    # strict.
    if onlyTwoSequences and len(rest):
        raise ValueError(f"Passed {len(rest)} unexpected extra sequences.")

    for read in r1, r2:
        if gapSymbol in read.sequence:
            raise ValueError(
                f"Sequence {read.id!r} contains one or more gap "
                f"characters {gapSymbol!r}."
            )

    alignment = edlib.getNiceAlignment(
        edlib.align(
            r1.sequence,
            r2.sequence,
            mode="NW",
            task="path",
            additionalEqualities=EDLIB_AMBIGUOUS if matchAmbiguous else None,
        ),
        r1.sequence,
        r2.sequence,
        gapSymbol=gapSymbol,
    )

    seq1, seq2 = alignment["query_aligned"], alignment["target_aligned"]

    # Try to remove unneeded gaps if requested. This is only possible if
    # the length of both sequences has changed.
    if minimizeGaps and len(seq1) != len(r1) and len(seq2) != len(r2):
        seq1, seq2 = removeUnnecessaryGaps(seq1, seq2, gapSymbol)

    return Reads([r1.__class__(r1.id, seq1), r2.__class__(r2.id, seq2)])



#!/usr/bin/env python

"""
Read simplified XML BLAST records and report the elapsed time.
"""

from Bio.Blast import NCBIXML
from time import time
import sys


if __name__ == "__main__":
    if len(sys.argv) != 2:
        print("Usage: %s file.json" % sys.argv[0], file=sys.stderr)
        sys.exit(1)
    else:
        start = time()
        with open(sys.argv[1]) as fp:
            for count, record in enumerate(NCBIXML.parse(fp)):
                pass
        elapsed = time() - start
        print(
            "Read %d XML BLAST records in %.3f secs (%.0f records/sec)"
            % (count, elapsed, float(count) / float(elapsed))
        )



import sys
import six
from unittest import TestCase, skipUnless
from six import assertRaisesRegex
from subprocess import CalledProcessError

from dark.process import Executor


class TestProcess(TestCase):
    """
    Test the Process class.
    """

    def testUnknownCommand(self):
        """
        An unknown command must raise CalledProcessError.
        """
        e = Executor()
        # Presumably there will not be an executable with this name!
        command = "/".join(["dev", "non-existent", "@" * 20])
        error = r"^Command '%s' returned non-zero exit status 127%s$" % (
            command,
            "" if sys.version_info < (3, 6) else r"\.",
        )
        assertRaisesRegex(
            self, CalledProcessError, error, e.execute, command, useStderr=False
        )

    def testDryRunTrue(self):
        """
        The dryRun attribute must be set when dryRun=True.
        """
        e = Executor(dryRun=True)
        self.assertTrue(e.dryRun)

    def testDryRunFalse(self):
        """
        The dryRun attribute must be set when dryRun=False.
        """
        e = Executor(dryRun=False)
        self.assertFalse(e.dryRun)

    def testDryRunDefault(self):
        e = Executor(dryRun=True)
        result = e.execute("date")
        self.assertIsNone(result)
        self.assertEqual("$ date", e.log[-1])

    def testDryRunDefaultOverride(self):
        """
        It must be possible to override the default dryRun setting by passing
        a value to C{execute}.
        """
        e = Executor(dryRun=False)
        result = e.execute("date", dryRun=True)
        self.assertIsNone(result)
        self.assertEqual("$ date", e.log[-1])

    @skipUnless(six.PY3, "subprocess output skipped under PY2")
    def testEchoStr(self):
        """
        We should be able to call echo using a string command.
        """
        e = Executor()
        result = e.execute("echo hello")
        self.assertEqual("hello", result.stdout.strip())
        self.assertTrue("$ echo hello" in e.log)

    @skipUnless(six.PY3, "subprocess output skipped under PY2")
    def testEchoList(self):
        """
        We should be able to call echo using a list command.
        """
        e = Executor()
        result = e.execute(["echo", "hello"])
        self.assertEqual("hello", result.stdout.strip())
        self.assertTrue("$ echo hello" in e.log)

    @skipUnless(six.PY3, "subprocess output skipped under PY2")
    def testPipe(self):
        """
        We should be able to pipe echo into wc -c.
        """
        e = Executor()
        result = e.execute("echo hello | wc -c")
        self.assertEqual("6", result.stdout.strip())
        self.assertTrue("$ echo hello | wc -c" in e.log)



from math import log
from collections import Counter


class ReadIntervals:
    """
    Hold information about the set of reads that match a subject.

    @param targetLength: The C{int} length of the target sequence that the
        reads are against.
    """

    EMPTY = 0
    FULL = 1

    def __init__(self, targetLength):
        self._targetLength = targetLength
        self._intervals = []

    def add(self, start, end):
        """
        Add the start and end offsets of a matching read.

        @param start: The C{int} start offset of the read match in the subject.
        @param end: The C{int} end offset of the read match in the subject.
            This is Python-style: the end offset is not included in the match.
        """
        assert start <= end
        self._intervals.append((start, end))

    def walk(self):
        """
        Get the non-overlapping read intervals that match the subject.

        @return: A generator that produces (TYPE, (START, END)) tuples, where
            where TYPE is either self.EMPTY or self.FULL and (START, STOP) is
            the interval. The endpoint (STOP) of the interval is not considered
            to be in the interval. I.e., the interval is really [START, STOP).
        """
        intervals = sorted(self._intervals)

        def nextFull():
            start, stop = intervals.pop(0)
            while intervals:
                if intervals[0][0] <= stop:
                    _, thisStop = intervals.pop(0)
                    if thisStop > stop:
                        stop = thisStop
                else:
                    break
            return (start, stop)

        if intervals:
            # If the first interval (read) starts after zero, yield an
            # initial empty section to get us to the first interval.
            if intervals[0][0] > 0:
                yield (self.EMPTY, (0, intervals[0][0]))

            while intervals:
                # Yield a full interval followed by an empty one (if there
                # is another interval pending).
                lastFull = nextFull()
                yield (self.FULL, lastFull)
                if intervals:
                    yield (self.EMPTY, (lastFull[1], intervals[0][0]))

            # Yield the final empty section, if any.
            if lastFull[1] < self._targetLength:
                yield (self.EMPTY, (lastFull[1], self._targetLength))

        else:
            yield (self.EMPTY, (0, self._targetLength))

    def coverage(self):
        """
        Get the fraction of a subject is matched by its set of reads.

        @return: The C{float} fraction of a subject matched by its reads.
        """
        if self._targetLength == 0:
            return 0.0

        coverage = 0
        for intervalType, (start, end) in self.walk():
            if intervalType == self.FULL:
                # Adjust start and end to ignore areas where the read falls
                # outside the target.
                coverage += min(end, self._targetLength) - max(0, start)
        return float(coverage) / self._targetLength

    def coverageCounts(self):
        """
        For each location in the subject, return a count of how many times that
        location is covered by a read.

        @return: a C{Counter} where the keys are the C{int} locations on the
            subject and the value is the number of times the location is
            covered by a read.
        """
        coverageCounts = Counter()
        for start, end in self._intervals:
            coverageCounts.update(range(max(0, start), min(self._targetLength, end)))
        return coverageCounts


class OffsetAdjuster:
    """
    A class that knows how to adjust the offsets in a normalized HSP according
    to the overall set of reads being plotted.

    @param intervals: An instance of L{ReadIntervals}.
    @param base: The C{float} logarithmic base to use when adjusting empty
        spaces in the hit sequence.
    """

    def __init__(self, intervals=None, base=2.0):
        self._adjustments = []  # Pairs of (X offset, adjustment).
        if intervals:
            divisor = log(base)
            for intervalType, (start, stop) in intervals.walk():
                if intervalType == ReadIntervals.EMPTY:
                    width = stop - start
                    logWidth = log(width) / divisor
                    self._adjustments.append((stop, width - logWidth))

    def adjustments(self):
        """
        Provide the adjustment values for this instance.

        @return: A C{list} of (X offset, adjustment) values, where the X
            offset is an C{int} and the adjustment is a C{float}.
        """
        return self._adjustments

    def _reductionForOffset(self, offset):
        """
        Calculate the total reduction for a given X axis offset.

        @param offset: The C{int} offset.
        @return: The total C{float} reduction that should be made for this
            offset.
        """
        reduction = 0
        for thisOffset, thisReduction in self._adjustments:
            if offset >= thisOffset:
                reduction += thisReduction
            else:
                break
        return reduction

    def adjustOffset(self, offset):
        """
        Adjust a single X offset.

        @param offset: The C{int} offset to adjust.
        @return: The C{float} adjusted offset.
        """
        return offset - self._reductionForOffset(offset)

    def adjustHSP(self, hsp):
        """
        Adjust the read and subject start and end offsets in an HSP.

        @param hsp: a L{dark.hsp.HSP} or L{dark.hsp.LSP} instance.
        """
        reduction = self._reductionForOffset(
            min(hsp.readStartInSubject, hsp.subjectStart)
        )

        hsp.readEndInSubject = hsp.readEndInSubject - reduction
        hsp.readStartInSubject = hsp.readStartInSubject - reduction
        hsp.subjectEnd = hsp.subjectEnd - reduction
        hsp.subjectStart = hsp.subjectStart - reduction



#!/usr/bin/env python

import sys
from collections import defaultdict
import csv
from operator import itemgetter
from typing import Union

from dark.dna import AMBIGUOUS
from dark.reads import addFASTACommandLineOptions, parseFASTACommandLineOptions

# These nucleotide codes will be listed first if --sortChars is used.
SPECIAL_CHARS = "ACGT-?N"


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(
        description="Print coverage information to standard output."
    )

    parser.add_argument(
        "--digits",
        default=4,
        type=int,
        help=(
            "The number of digits of precision to use for floating point "
            "outputs (coverage, gc fraction)."
        ),
    )

    parser.add_argument(
        "--noNames",
        dest="printNames",
        action="store_false",
        help="Do not print the ids of input sequences.",
    )

    parser.add_argument(
        "--strict",
        action="store_true",
        help=(
            "Treat all ambiguous nucleotides as not covered. Additional "
            "characters specified with --chars will also be counted as not "
            "covered."
        ),
    )

    parser.add_argument(
        "--gc", action="store_true", help="Include the GC percentage in the output."
    )

    parser.add_argument(
        "--reverse", "-r", action="store_true", help="Sort lines by decreasing value."
    )

    parser.add_argument(
        "--sortBy",
        help=(
            "What to sort lines by. Either one of id, coverage, length, gc, "
            "not-covered or else a single nucleotide character."
        ),
    )

    parser.add_argument(
        "--header",
        action="store_true",
        help="Write a header line and in cells print just the values.",
    )

    parser.add_argument("--sep", default="\t", help="The output separator character.")

    parser.add_argument(
        "--printCounts",
        action="store_true",
        help="Print the count for each character found in each read.",
    )

    parser.add_argument(
        "--sortChars", action="store_true", help="Sort the output columns."
    )

    addFASTACommandLineOptions(parser)
    args = parser.parse_args()
    reads = parseFASTACommandLineOptions(args)
    coveredChars = set("AGCT") if args.strict else set(AMBIGUOUS)
    data = []
    sortBy = args.sortBy
    keepCounts = args.printCounts or args.gc or (sortBy and len(sortBy) == 1)

    for read in reads:
        notCoveredCount = 0
        if keepCounts:
            counts: dict[str, int] = defaultdict(int)
        for char in read.sequence.upper():
            notCoveredCount += char not in coveredChars
            if keepCounts:
                counts[char] += 1

        length = len(read)

        # Sanity check.
        if keepCounts:
            assert sum(counts.values()) == length

        coverage = round(
            (length - notCoveredCount) / length if length else 0.0, args.digits
        )

        readData = {
            "id": read.id,
            "length": len(read),
            "not-covered": notCoveredCount,
            "coverage": coverage,
        }

        if args.printCounts:
            readData["counts"] = counts

        if args.gc:
            gc = 0
            for char in "GC":
                if char in counts:
                    gc += counts[char]
            readData["gc"] = round(gc / length, args.digits)

        data.append(readData)

    if not data:
        sys.exit(0)

    writerow = csv.writer(sys.stdout, delimiter=args.sep).writerow

    if args.sortChars:
        allChars = set()
        for d in data:
            allChars.update(d["counts"])

        specials = [char for char in SPECIAL_CHARS if char in allChars]
        allCharsList = specials + sorted(allChars - set(specials))

    header = args.header

    if header:
        hline = ["Id"] if args.printNames else []
        hline.extend(("Length", "Not covered", "Coverage"))
        if args.gc:
            hline.append("GC")
        if args.printCounts and args.sortChars:
            hline.extend(allCharsList)
        writerow(hline)

    if sortBy:
        if sortBy in data[0]:
            key = itemgetter(sortBy)  # type: ignore
        else:
            if len(sortBy) > 1:
                print("Cannot sort by unknown column {sortBy!r}.", file=sys.stderr)
                sys.exit(1)

            sortBy = sortBy.upper()

            def key(d):  # type: ignore
                return d["counts"].get(sortBy, 0)

        data = sorted(data, key=key, reverse=args.reverse)

    for d in data:
        line: list[Union[str, int]] = [d["id"]] if args.printNames else []
        line.append(d["length"] if header else f"length:{d['length']}")
        line.append(d["not-covered"] if header else f"not-covered:{d['not-covered']}")
        line.append(d["coverage"] if header else f"coverage:{d['coverage']}")
        if args.gc:
            line.append(d["gc"] if header else f"gc:{d['gc']}")

        if args.printCounts:
            counts = d["counts"]
            if args.sortChars:
                for char in allCharsList:
                    if char in counts:
                        line.append(
                            counts[char] if header else f"{char}:{counts[char]}"
                        )
                    else:
                        line.append("")
            else:
                for char, count in sorted(counts.items()):
                    line.append(count if header else f"{char}:{count}")

        writerow(line)



from unittest import TestCase
from pysam import CSOFT_CLIP, CHARD_CLIP, CMATCH, CINS

from dark.cigar import (
    CINS_STR,
    CDEL_STR,
    CMATCH_STR,
    CEQUAL_STR,
    CDIFF_STR,
    dna2cigar,
    makeCigar,
    cigarTuplesToOperations,
    softClippedOffset,
    insertionOffset,
)


class TestConstants(TestCase):
    """
    Test constants in dark.cigar
    """

    def testOperations(self):
        """
        Make sure the CIGAR operation strings have the expected one-letter
        codes.
        """
        self.assertEqual("I", CINS_STR)
        self.assertEqual("D", CDEL_STR)
        self.assertEqual("M", CMATCH_STR)
        self.assertEqual("=", CEQUAL_STR)
        self.assertEqual("X", CDIFF_STR)


class TestDNA2CIGAR(TestCase):
    """
    Test the dna2cigar function.
    """

    def testUnequalStrings(self):
        """
        If two unequal strings are passed, a ValueError must be raised.
        """
        error = r"^Sequences 'hey' and 'there' of unequal length \(3 != 5\)\.$"
        self.assertRaisesRegex(ValueError, error, dna2cigar, "hey", "there")

    def testEmptyStrings(self):
        """
        If two empty strings are passed, a ValueError must be raised.
        """
        error = r"^Two sequences of zero length were passed\.$"
        self.assertRaisesRegex(ValueError, error, dna2cigar, "", "")

    def testEqualStringsLength1(self):
        """
        If two equal strings of length 1 are passed, the correct CIGAR string
        must be returned.
        """
        self.assertEqual("1=", dna2cigar("A", "A"))

    def testEqualStringsLength1MismatchedCase(self):
        """
        If two strings of length 1 are passed and the're the same but for
        case, the CIGAR string must indicate that they do not match.
        """
        self.assertEqual("1X", dna2cigar("A", "a"))

    def testEqualStringsLength1Concise(self):
        """
        If two equal strings of length 1 are passed and the concise argument
        is True, the correct CIGAR string must be returned.
        """
        self.assertEqual("1M", dna2cigar("A", "A", concise=True))

    def testUnequalStringsLength1(self):
        """
        If two unequal strings of length 1 are passed, the correct CIGAR
        string must be returned.
        """
        self.assertEqual("1X", dna2cigar("A", "G"))

    def testUnequalStringsLength1Concise(self):
        """
        If two unequal strings of length 1 are passed and the concise argument
        is True, the correct CIGAR string must be returned.
        """
        self.assertEqual("1M", dna2cigar("A", "G", concise=True))

    def testEqualStringsLength2(self):
        """
        If two equal strings of length 2 are passed, the correct CIGAR string
        must be returned.
        """
        self.assertEqual("2=", dna2cigar("AT", "AT"))

    def testEqualStringsLength2Concise(self):
        """
        If two equal strings of length 2 are passed and the concise argument
        is True, the correct CIGAR string must be returned.
        """
        self.assertEqual("2M", dna2cigar("AT", "AT", concise=True))

    def testUnequalStringsLength2(self):
        """
        If two unequal strings of length 2 are passed, the correct CIGAR
        string must be returned.
        """
        self.assertEqual("2X", dna2cigar("AT", "GC"))

    def testUnequalStringsLength2Concise(self):
        """
        If two unequal strings of length 2 are passed and the concise argument
        is True, the correct CIGAR string must be returned.
        """
        self.assertEqual("2M", dna2cigar("AT", "GC", concise=True))

    def testMixedEqualityStringsLength2(self):
        """
        If two strings of length 2 are passed and they match in just one
        place, the correct CIGAR string must be returned.
        """
        self.assertEqual("1=1X", dna2cigar("AG", "AT"))

    def testMixedEqualityStringsLength2Concise(self):
        """
        If two equal strings of length 2 are passed and the concise argument
        is True, the correct CIGAR string must be returned.
        """
        self.assertEqual("2M", dna2cigar("AG", "AT", concise=True))

    def testMixedEqualityStringsLength2Alt(self):
        """
        If two strings of length 2 are passed and they match in just one
        place, the correct CIGAR string must be returned.
        """
        self.assertEqual("1X1=", dna2cigar("GA", "TA"))

    def testMixedEqualityStringsLength2ConciseAlt(self):
        """
        If two equal strings of length 2 are passed and the concise argument
        is True, the correct CIGAR string must be returned.
        """
        self.assertEqual("2M", dna2cigar("GA", "TA", concise=True))

    def testMixedMatch(self):
        """
        If two strings with matching and non-matching regions must result
        in the expected CIGAR string.
        """
        self.assertEqual("3=4X5=2X", dna2cigar("ACGTTTTCCCTTGG", "ACGAAAACCCTTTT"))

    def testMixedMatchConcise(self):
        """
        If two strings with matching and non-matching regions must result
        in the expected CIGAR string when the concise argument is True.
        """
        self.assertEqual(
            "14M", dna2cigar("ACGTTTTCCCTTGG", "ACGAAAACCCTTTT", concise=True)
        )


class TestMakeCigar(TestCase):
    """
    Test making CIGAR strings.
    """

    def testEmptyReference(self):
        """
        If the reference is empty, a ValueError must be raised.
        """
        error = r"^Empty reference$"
        self.assertRaisesRegex(ValueError, error, makeCigar, "", "ACGT")

    def testEmptyQuery(self):
        """
        If the query is empty, a ValueError must be raised.
        """
        error = r"^Empty query$"
        self.assertRaisesRegex(ValueError, error, makeCigar, "ACGT", "")

    def testOneBaseInitialMatch(self):
        """
        If the query has one base that matches the start of the reference, 1M
        should be the result.
        """
        self.assertEqual("1M", makeCigar("ACGT", "A"))

    def testTwoBasesInitialMatch(self):
        """
        If the query has two bases that matches the start of the reference,
        1M1M should be the result.
        """
        self.assertEqual("2M", makeCigar("ACGT", "AT"))

    def testOneBaseFinalMatch(self):
        """
        If the query has one base that matches the end of the reference, 1M
        should be the result.
        """
        self.assertEqual("1M", makeCigar("ACGT", "   A"))

    def testTwoBasesFinalMatch(self):
        """
        If the query has two bases that matches the end of the reference, 1M1M
        should be the result.
        """
        self.assertEqual("2M", makeCigar("ACGT", "  AT"))

    def testEqualStrings(self):
        """
        If the query exactly overlaps the reference we should get 'M' as many
        times as the length of the sequences.
        """
        self.assertEqual("4M", makeCigar("ACGT", "ATGG"))

    def testLeftClippingOne(self):
        """
        Test that left soft clipping of one base works.
        """
        self.assertEqual("1S2M", makeCigar(" ACGT", "TAT"))

    def testLeftClippingTwo(self):
        """
        Test that left soft clipping of two bases works.
        """
        self.assertEqual("2S1M", makeCigar("  ACGT", "TCA"))

    def testRightClippingUnpaddedOne(self):
        """
        Test that right soft clipping of one base works when the reference
        does not have enough padding to match the length of the query.
        """
        self.assertEqual("2M1S", makeCigar("  ACGT", "    ATA"))

    def testRightClippingUnpaddedTwo(self):
        """
        Test that right soft clipping of two bases works when the reference
        does not have enough padding to match the length of the query.
        """
        self.assertEqual("2M2S", makeCigar("  ACGT", "    ATAA"))

    def testRightClippingPaddedOne(self):
        """
        Test that right soft clipping of one base works when the reference
        is padded on the right.
        """
        self.assertEqual("2M1S", makeCigar("  ACGT ", "    ATA"))

    def testRightClippingPaddedTwo(self):
        """
        Test that right soft clipping of two bases works when the reference
        is padded on the right.
        """
        self.assertEqual("2M2S", makeCigar("  ACGT  ", "    ATAA"))

    def testLeftInsert(self):
        """
        Test that an insert on the left works.
        """
        self.assertEqual("2I2M", makeCigar("--ACGT  ", "ATAC", noEdgeInsertions=False))

        # When edge insertions are not allowed, the first two bases are
        # soft-clipped (as in tests above).
        self.assertEqual("2S2M", makeCigar("--ACGT  ", "ATAC"))


class TestCigarTuplesToOperations(TestCase):
    """
    Test the cigarTuplesToOperations function.
    """

    def testEmpty(self):
        """
        If there are no tuples, the empty list must be returned.
        """
        self.assertEqual([], list(cigarTuplesToOperations([])))

    def testOneOfOne(self):
        """
        Test one tuple with one operation of length one.
        """
        self.assertEqual([CINS], list(cigarTuplesToOperations([(CINS, 1)])))

    def testOneOfTwo(self):
        """
        Test one tuple with one operation of length two.
        """
        self.assertEqual([CINS, CINS], list(cigarTuplesToOperations([(CINS, 2)])))

    def testTwo(self):
        """
        Test two tuples.
        """
        self.assertEqual(
            [CINS, CINS, CMATCH, CMATCH, CMATCH],
            list(cigarTuplesToOperations([(CINS, 2), (CMATCH, 3)])),
        )

    def testHardClippingIncluded(self):
        """
        Hard clipping operations should be included.
        """
        self.assertEqual(
            [CHARD_CLIP, CHARD_CLIP, CINS, CINS, CMATCH, CMATCH, CMATCH],
            list(cigarTuplesToOperations([(CHARD_CLIP, 2), (CINS, 2), (CMATCH, 3)])),
        )

    def testSkipHardClipping(self):
        """
        It must be possible to have hard clipping operations skipped.
        """
        self.assertEqual(
            [CINS, CINS, CMATCH, CMATCH, CMATCH],
            list(
                cigarTuplesToOperations(
                    [(CHARD_CLIP, 2), (CINS, 2), (CMATCH, 3)], includeHardClip=False
                )
            ),
        )


class TestSoftClippedOffset(TestCase):
    """
    Test the softClippedOffset function.
    """

    def testNoNonClippedBases(self):
        """
        If there are no preceding or following non-clipped bases, a ValueError
        must be raised.
        """
        error = (
            r"^Soft-clipped base with no following or preceding "
            r"non-hard-clipped bases\.$"
        )
        self.assertRaisesRegex(
            ValueError, error, softClippedOffset, 0, ((0, None),), (CSOFT_CLIP,)
        )

    def testMatchOneBefore(self):
        """
        Test that a soft-clipped base one site before a non-soft-clipped site
        returns the correct offset.
        """
        self.assertEqual(
            9, softClippedOffset(0, ((0, None), (1, 10)), (CSOFT_CLIP, CMATCH))
        )

    def testMatchTwoBefore(self):
        """
        Test that a soft-clipped base two sites before a non-soft-clipped site
        returns the correct offset.
        """
        self.assertEqual(
            8,
            softClippedOffset(
                0, ((0, None), (1, None), (2, 10)), (CSOFT_CLIP, CSOFT_CLIP, CMATCH)
            ),
        )

    def testMatchTwoBeforeWithHardClips(self):
        """
        Test that a soft-clipped base two sites before a non-soft-clipped site
        returns the correct offset.
        """
        self.assertEqual(
            8,
            softClippedOffset(
                2,
                ((0, None), (1, None), (2, None), (3, None), (4, 10)),
                (CHARD_CLIP, CHARD_CLIP, CSOFT_CLIP, CSOFT_CLIP, CMATCH),
            ),
        )

    def testMatchOneAfter(self):
        """
        Test that a soft-clipped base one site after a non-soft-clipped site
        returns the correct offset.
        """
        self.assertEqual(
            11, softClippedOffset(1, ((0, 10), (1, None)), (CMATCH, CSOFT_CLIP))
        )

    def testMatchTwoAfter(self):
        """
        Test that a soft-clipped base two sites after a non-soft-clipped site
        returns the correct offset.
        """
        self.assertEqual(
            12,
            softClippedOffset(
                2, ((0, 10), (1, None), (2, None)), (CMATCH, CSOFT_CLIP, CSOFT_CLIP)
            ),
        )

    def testMatchTwoAfterThenHardClips(self):
        """
        Test that a soft-clipped base two sites after a non-soft-clipped site
        returns the correct offset, including when there are also hard clips.
        """
        self.assertEqual(
            12,
            softClippedOffset(
                2,
                ((0, 10), (1, None), (2, None), (3, None), (4, None)),
                (CMATCH, CSOFT_CLIP, CSOFT_CLIP, CHARD_CLIP, CHARD_CLIP),
            ),
        )


class TestInsertionOffset(TestCase):
    """
    Test the insertionOffset function.
    """

    def testNoNonInsertionBases(self):
        """
        If there are no preceding or following non-inserted bases, a ValueError
        must be raised.
        """
        error = r"^Inserted base with no following or preceding reference " r"bases\.$"
        self.assertRaisesRegex(
            ValueError, error, insertionOffset, 0, ((0, None),), (CINS,)
        )

    def testMatchOneBefore(self):
        """
        Test that an inserted base one site before a non-inserted site
        returns the correct offset.
        """
        self.assertEqual(
            (False, 10), insertionOffset(0, ((0, None), (1, 10)), (CINS, CMATCH))
        )

    def testMatchTwoBefore(self):
        """
        Test that an inserted base two sites before a non-inserted site
        returns the correct offset.
        """
        self.assertEqual(
            (False, 10),
            insertionOffset(0, ((0, None), (1, None), (2, 10)), (CINS, CINS, CMATCH)),
        )

    def testMatchTwoBeforeWithHardClips(self):
        """
        Test that an inserted base two sites before a non-inserted site
        returns the correct offset.
        """
        self.assertEqual(
            (False, 10),
            insertionOffset(
                2,
                ((0, None), (1, None), (2, None), (3, None), (4, 10)),
                (CHARD_CLIP, CHARD_CLIP, CINS, CINS, CMATCH),
            ),
        )

    def testMatchOneAfter(self):
        """
        Test that an inserted base one site after a non-inserted site
        returns the correct offset.
        """
        self.assertEqual(
            (True, 11), insertionOffset(1, ((0, 10), (1, None)), (CMATCH, CINS))
        )

    def testMatchTwoAfter(self):
        """
        Test that an inserted base two sites after a non-inserted site
        returns the correct offset.
        """
        self.assertEqual(
            (True, 11),
            insertionOffset(2, ((0, 10), (1, None), (2, None)), (CMATCH, CINS, CINS)),
        )

    def testMatchTwoAfterThenHardClips(self):
        """
        Test that an inserted base two sites after a non-inserted site
        returns the correct offset, including when there are also hard clips.
        """
        self.assertEqual(
            (True, 11),
            insertionOffset(
                2,
                ((0, 10), (1, None), (2, None), (3, None), (4, None)),
                (CMATCH, CINS, CINS, CHARD_CLIP, CHARD_CLIP),
            ),
        )



from unittest import TestCase, skip
from unittest.mock import patch
from six import assertRaisesRegex
from six.moves import builtins
from contextlib import contextmanager
from io import StringIO

from dark.proteins import (
    splitNames,
    _NO_PATHOGEN_NAME,
    getPathogenProteinCounts,
    ProteinGrouper,
    PathogenSampleFiles,
)


class TestSplitNames(TestCase):
    """
    Tests for the splitNames function.
    """

    def testNoBrackets(self):
        """
        If a string with no trailing pathogen name in square brackets is given,
        splitNames must return its argument followed by a string indicating
        no pathogen name could be found.
        """
        self.assertEqual(("xxx", _NO_PATHOGEN_NAME), splitNames("xxx"))

    def testTwoSetsOfBrackets(self):
        """
        If a string with two trailing substrings in square brackets is given,
        splitNames must extract the substring from the second set of square
        brackets and use that as the pathogen name.
        """
        self.assertEqual(
            ("xxx [other]", "pathogen"), splitNames("xxx [other] [pathogen]")
        )

    def testWhitespaceStripping(self):
        """
        If a string with names that have whitespace is passed, splitNames
        must strip the whitespace in its result.
        """
        self.assertEqual(("xxx", "pathogen"), splitNames(" xxx [ pathogen ]"))

    def testNestedBrackets(self):
        """
        If a string with two nested trailing substrings in square brackets is
        given, splitNames must return its argument followed by a string
        indicating no pathogen name could be found.
        """
        self.assertEqual(
            ("xxx [nested [pathogen name]]", _NO_PATHOGEN_NAME),
            splitNames("xxx [nested [pathogen name]]"),
        )

    def testNormalCase(self):
        """
        If a string with a protein and pathogen name is passed, splitNames
        must return the expected result.
        """
        self.assertEqual(
            ("protein name", "pathogen name"),
            splitNames("protein name [pathogen name]"),
        )


class TestGetPathogenProteinCounts(TestCase):
    """
    Tests for the getPathogenProteinCounts function.
    """

    def testNone(self):
        """
        getPathogenProteinCounts must return an empty result if passed None as
        the protein FASTA file.
        """
        self.assertEqual({}, getPathogenProteinCounts(None))

    def testExpected(self):
        """
        getPathogenProteinCounts must return the expected result.
        """

        class SideEffect:
            def __init__(self, test):
                self.test = test
                self.count = 0

            def sideEffect(self, filename, **kwargs):
                if self.count == 0:
                    self.test.assertEqual("filename.fasta", filename)
                    self.count += 1
                    return StringIO(
                        ">protein 1 [pathogen 1]\n"
                        + "ACTG\n"
                        + ">protein 2 [pathogen 1]\n"
                        + "AA\n"
                        + ">no pathogen name here\n"
                        + "AA\n"
                        + ">protein 3 [pathogen 2]\n"
                        + "AA\n"
                    )
                else:
                    self.test.fail("We are only supposed to be called once!")

        sideEffect = SideEffect(self)
        with patch.object(builtins, "open") as mockMethod:
            mockMethod.side_effect = sideEffect.sideEffect
            self.assertEqual(
                {
                    "pathogen 1": 2,
                    "pathogen 2": 1,
                },
                getPathogenProteinCounts(["filename.fasta"]),
            )
            self.assertEqual(1, sideEffect.count)

    def testExpectedWithTwoFiles(self):
        """
        getPathogenProteinCounts must return the expected result when details
        are read from two FASTA files.
        """

        class SideEffect:
            def __init__(self, test):
                self.test = test
                self.count = 0

            def sideEffect(self, filename, **kwargs):
                if self.count == 0:
                    self.test.assertEqual("filename1.fasta", filename)
                    self.count += 1
                    return StringIO(
                        ">protein 1 [pathogen 1]\n"
                        + "ACTG\n"
                        + ">protein 3 [pathogen 2]\n"
                        + "AA\n"
                    )
                elif self.count == 1:
                    self.test.assertEqual("filename2.fasta", filename)
                    self.count += 1
                    return StringIO(">protein 2 [pathogen 1]\n" + "AA\n")
                else:
                    self.test.fail("We are only supposed to be called twice!")

        sideEffect = SideEffect(self)
        with patch.object(builtins, "open") as mockMethod:
            mockMethod.side_effect = sideEffect.sideEffect
            self.assertEqual(
                {
                    "pathogen 1": 2,
                    "pathogen 2": 1,
                },
                getPathogenProteinCounts(["filename1.fasta", "filename2.fasta"]),
            )
            self.assertEqual(2, sideEffect.count)


class TestProteinGrouper(TestCase):
    """
    Tests for the dark.proteins.ProteinGrouper class.
    """

    def testUnknownFormat(self):
        """
        Passing an unknown format argument must result in a ValueError
        being raised.
        """
        error = "^format_ must be either 'fasta' or 'fastq'\\.$"
        assertRaisesRegex(self, ValueError, error, ProteinGrouper, format_="unknown")

    def testNoAssetDir(self):
        """
        If no asset directorey is given to a protein grouper, its _assetDir
        attribute be the default ('out').
        """
        pg = ProteinGrouper()
        self.assertEqual("out", pg._assetDir)

    def testAssetDir(self):
        """
        If an asset directorey is given to a protein grouper, its _assetDir
        attribute be set to hold that value.
        """
        pg = ProteinGrouper(assetDir="xxx")
        self.assertEqual("xxx", pg._assetDir)

    def testNoSampleName(self):
        """
        If no sample name is given to a protein grouper, its _sampleName
        attribute must be None.
        """
        pg = ProteinGrouper()
        self.assertEqual(None, pg._sampleName)

    def testNoRegex(self):
        """
        If no regex is given to a protein grouper, its _sampleNameRegex
        attribute mustbe None.
        """
        pg = ProteinGrouper()
        self.assertEqual(None, pg._sampleNameRegex)

    def testNoFiles(self):
        """
        If no files have been given to a protein grouper, its sample names and
        pathogen names attributes must both be empty.
        """
        pg = ProteinGrouper()
        self.assertEqual({}, pg.pathogenNames)
        self.assertEqual({}, pg.sampleNames)

    def testUnknownPathogenType(self):
        """
        If the toHTML method of a protein grouper is given an unknown pathogen
        type it must raise a ValueError.
        """
        pg = ProteinGrouper()
        error = (
            "^Unrecognized pathogenType argument: 'x'\\. Value must be "
            "either 'bacterial' or 'viral'\\.$"
        )
        assertRaisesRegex(self, ValueError, error, pg.toHTML, pathogenType="x")

    def testDuplicatePathogenProteinSample(self):
        """
        If a protein grouper is given duplicate information for a
        pathogen/protein/sample combination it must raise a ValueError.
        """
        fp = StringIO(
            "0.77 46.6 48.1 5 6 74 acc|GENBANK|I44.6|GENBANK|J77|"
            "ubiquitin [Lausannevirus]\n"
        )
        pg = ProteinGrouper()
        pg.addFile("sample", fp)
        fp.seek(0)
        error = (
            "^Protein 'acc\\|GENBANK\\|I44.6\\|GENBANK\\|J77|"
            "ubiquitin' already seen for pathogen 'Lausannevirus' "
            "sample 'sample'\\.$"
        )
        assertRaisesRegex(self, ValueError, error, pg.addFile, "sample", fp)

    def testOneLineInOneFile(self):
        """
        If a protein grouper is given one file with one line, its pathogenNames
        dict must be as expected.
        """
        fp = StringIO(
            "0.77 46.6 48.1 5 6 74 acc|GENBANK|I44.6|GENBANK|J77|"
            "ubiquitin [Lausannevirus]\n"
        )
        pg = ProteinGrouper()
        pg.addFile("sample-filename", fp)
        self.assertEqual(
            {
                "Lausannevirus": {
                    "sample-filename": {
                        "proteins": {
                            "acc|GENBANK|I44.6|GENBANK|J77|ubiquitin": {
                                "bestScore": 48.1,
                                "bluePlotFilename": "out/0.png",
                                "coverage": 0.77,
                                "readsFilename": "out/0.fasta",
                                "hspCount": 6,
                                "index": 0,
                                "medianScore": 46.6,
                                "outDir": "out",
                                "proteinLength": 74,
                                "proteinName": (
                                    "acc|GENBANK|I44.6|GENBANK|" "J77|ubiquitin"
                                ),
                                "proteinURL": (
                                    "http://www.ncbi.nlm.nih.gov/nuccore/" "I44.6"
                                ),
                                "genomeURL": (
                                    "http://www.ncbi.nlm.nih.gov/nuccore/J77"
                                ),
                                "readCount": 5,
                                "readAndHspCountStr": "5/6",
                            },
                        },
                        "uniqueReadCount": None,
                    },
                }
            },
            pg.pathogenNames,
        )

    def testOneLineInOneFileWithDifferentAssetDir(self):
        """
        If a protein grouper is given a different assetDir name,
        the outDir needs to have that same name, as expected.
        """
        fp = StringIO(
            "0.77 46.6 48.1 5 6 74 acc|GENBANK|I44.6|GENBANK|J77|"
            "ubiquitin [Lausannevirus]\n"
        )
        pg = ProteinGrouper(assetDir="differentname")
        pg.addFile("sample-filename", fp)
        self.assertEqual(
            {
                "Lausannevirus": {
                    "sample-filename": {
                        "proteins": {
                            "acc|GENBANK|I44.6|GENBANK|J77|ubiquitin": {
                                "bestScore": 48.1,
                                "bluePlotFilename": "differentname/0.png",
                                "coverage": 0.77,
                                "readsFilename": "differentname/0.fasta",
                                "hspCount": 6,
                                "index": 0,
                                "medianScore": 46.6,
                                "outDir": "differentname",
                                "proteinLength": 74,
                                "proteinName": (
                                    "acc|GENBANK|I44.6|GENBANK|" "J77|ubiquitin"
                                ),
                                "proteinURL": (
                                    "http://www.ncbi.nlm.nih.gov/" "nuccore/I44.6"
                                ),
                                "genomeURL": (
                                    "http://www.ncbi.nlm.nih.gov/nuccore/J77"
                                ),
                                "readCount": 5,
                                "readAndHspCountStr": "5/6",
                            },
                        },
                        "uniqueReadCount": None,
                    },
                }
            },
            pg.pathogenNames,
        )

    def testOneLineInOneFileFASTQ(self):
        """
        If a protein grouper is given one file with one line, its pathogenNames
        dict must be as expected, including for a FASTQ file.
        """
        fp = StringIO(
            "0.77 46.6 48.1 5 6 74 acc|GENBANK|I44.6|GENBANK|J77|"
            "ubiquitin [Lausannevirus]\n"
        )
        pg = ProteinGrouper(format_="fastq")
        pg.addFile("sample-filename", fp)
        self.assertEqual(
            {
                "Lausannevirus": {
                    "sample-filename": {
                        "proteins": {
                            "acc|GENBANK|I44.6|GENBANK|J77|ubiquitin": {
                                "bestScore": 48.1,
                                "bluePlotFilename": "out/0.png",
                                "coverage": 0.77,
                                "readsFilename": "out/0.fastq",
                                "hspCount": 6,
                                "index": 0,
                                "medianScore": 46.6,
                                "outDir": "out",
                                "proteinLength": 74,
                                "proteinName": (
                                    "acc|GENBANK|I44.6|GENBANK|" "J77|ubiquitin"
                                ),
                                "proteinURL": (
                                    "http://www.ncbi.nlm.nih.gov/" "nuccore/I44.6"
                                ),
                                "genomeURL": (
                                    "http://www.ncbi.nlm.nih.gov/nuccore/J77"
                                ),
                                "readCount": 5,
                                "readAndHspCountStr": "5/6",
                            },
                        },
                        "uniqueReadCount": None,
                    },
                }
            },
            pg.pathogenNames,
        )

    def testOneLineInOneFileTitle(self):
        """
        If a protein grouper is given one file with one line, its _title method
        must return the expected string.
        """
        fp = StringIO(
            "0.77 46.6 48.1 5 6 74 acc|GENBANK|I44.6|GENBANK|J77|"
            "ubiquitin [Lausannevirus]\n"
        )
        pg = ProteinGrouper()
        pg.addFile("sample-filename", fp)
        self.assertEqual(
            "Overall, proteins from 1 pathogen were found in 1 sample.", pg._title()
        )

    def testTwoLinesInOneFileTitle(self):
        """
        If a protein grouper is given one file with two protein lines, each
        from a different pathogen, its _title method must return the expected
        string.
        """
        fp = StringIO(
            "0.77 46.6 48.1 5 6 74 acc|GENBANK|I44.6|GENBANK|J77|"
            "ubiquitin [Lausannevirus]\n"
            "0.77 46.6 48.1 5 6 74 acc|GENBANK|I44.6|GENBANK|J77|"
            "ubiquitin [X Virus]\n"
        )
        pg = ProteinGrouper()
        pg.addFile("sample-filename", fp)
        self.assertEqual(
            "Overall, proteins from 2 pathogens were found in 1 sample.", pg._title()
        )

    def testTwoLinesInOneFileSamePathogen(self):
        """
        If a protein grouper is given one file with two lines from the same
        pathogen, its pathogenNames dict must be as expected.
        """
        fp = StringIO(
            "0.63 41.3 44.2 9 9 12 acc|GENBANK|I44.6|GENBANK|J77|VP1 "
            "[Lausannevirus]\n"
            "0.77 46.6 48.1 5 6 74 acc|GENBANK|I44.7|GENBANK|J78|VP2 "
            "[Lausannevirus]\n"
        )
        pg = ProteinGrouper()
        pg.addFile("sample-filename", fp)
        self.assertEqual(
            {
                "Lausannevirus": {
                    "sample-filename": {
                        "proteins": {
                            "acc|GENBANK|I44.6|GENBANK|J77|VP1": {
                                "bestScore": 44.2,
                                "bluePlotFilename": "out/0.png",
                                "coverage": 0.63,
                                "readsFilename": "out/0.fasta",
                                "hspCount": 9,
                                "index": 0,
                                "medianScore": 41.3,
                                "outDir": "out",
                                "proteinLength": 12,
                                "proteinName": ("acc|GENBANK|I44.6|GENBANK|" "J77|VP1"),
                                "proteinURL": (
                                    "http://www.ncbi.nlm.nih.gov/" "nuccore/I44.6"
                                ),
                                "genomeURL": (
                                    "http://www.ncbi.nlm.nih.gov/nuccore/J77"
                                ),
                                "readCount": 9,
                                "readAndHspCountStr": "9",
                            },
                            "acc|GENBANK|I44.7|GENBANK|J78|VP2": {
                                "bestScore": 48.1,
                                "bluePlotFilename": "out/1.png",
                                "coverage": 0.77,
                                "readsFilename": "out/1.fasta",
                                "hspCount": 6,
                                "index": 1,
                                "medianScore": 46.6,
                                "outDir": "out",
                                "proteinLength": 74,
                                "proteinName": ("acc|GENBANK|I44.7|GENBANK|" "J78|VP2"),
                                "proteinURL": (
                                    "http://www.ncbi.nlm.nih.gov/" "nuccore/I44.7"
                                ),
                                "genomeURL": (
                                    "http://www.ncbi.nlm.nih.gov/nuccore/J78"
                                ),
                                "readCount": 5,
                                "readAndHspCountStr": "5/6",
                            },
                        },
                        "uniqueReadCount": None,
                    },
                },
            },
            pg.pathogenNames,
        )

    def testTwoLinesInOneFileDifferentPathogens(self):
        """
        If a protein grouper is given one file with two lines from different
        pathogens, its pathogenNames dict must be as expected.
        """
        fp = StringIO(
            "0.63 41.3 44.2 9 9 12 acc|GENBANK|I44.6|GENBANK|J77|VP1 "
            "[Lausannevirus]\n"
            "0.77 46.6 48.1 5 6 74 acc|GENBANK|I44.7|GENBANK|J78|VP2 "
            "[Hepatitis B virus]\n"
        )
        pg = ProteinGrouper()
        pg.addFile("sample-filename", fp)
        self.assertEqual(
            {
                "Lausannevirus": {
                    "sample-filename": {
                        "proteins": {
                            "acc|GENBANK|I44.6|GENBANK|J77|VP1": {
                                "bestScore": 44.2,
                                "bluePlotFilename": "out/0.png",
                                "coverage": 0.63,
                                "readsFilename": "out/0.fasta",
                                "hspCount": 9,
                                "index": 0,
                                "medianScore": 41.3,
                                "outDir": "out",
                                "proteinLength": 12,
                                "proteinName": ("acc|GENBANK|I44.6|GENBANK|J77|VP1"),
                                "proteinURL": (
                                    "http://www.ncbi.nlm.nih.gov/" "nuccore/I44.6"
                                ),
                                "genomeURL": (
                                    "http://www.ncbi.nlm.nih.gov/nuccore/J77"
                                ),
                                "readCount": 9,
                                "readAndHspCountStr": "9",
                            },
                        },
                        "uniqueReadCount": None,
                    },
                },
                "Hepatitis B virus": {
                    "sample-filename": {
                        "proteins": {
                            "acc|GENBANK|I44.7|GENBANK|J78|VP2": {
                                "bestScore": 48.1,
                                "bluePlotFilename": "out/1.png",
                                "coverage": 0.77,
                                "readsFilename": "out/1.fasta",
                                "hspCount": 6,
                                "index": 1,
                                "medianScore": 46.6,
                                "outDir": "out",
                                "proteinLength": 74,
                                "proteinName": ("acc|GENBANK|I44.7|GENBANK|J78|VP2"),
                                "proteinURL": (
                                    "http://www.ncbi.nlm.nih.gov/" "nuccore/I44.7"
                                ),
                                "genomeURL": (
                                    "http://www.ncbi.nlm.nih.gov/nuccore/J78"
                                ),
                                "readCount": 5,
                                "readAndHspCountStr": "5/6",
                            },
                        },
                        "uniqueReadCount": None,
                    },
                },
            },
            pg.pathogenNames,
        )

    def testOneLineInEachOfTwoFilesSamePathogen(self):
        """
        If a protein grouper is given two files, each with one line from the
        same pathogen, its pathogenNames dict must be as expected.
        """
        fp1 = StringIO(
            "0.63 41.3 44.2 9 9 12 acc|GENBANK|I44.6|GENBANK|J77|VP1 "
            "[Lausannevirus]\n"
        )
        fp2 = StringIO(
            "0.77 46.6 48.1 5 6 74 acc|GENBANK|I44.7|GENBANK|J78|VP2 "
            "[Lausannevirus]\n"
        )
        pg = ProteinGrouper()
        pg.addFile("sample-filename-1", fp1)
        pg.addFile("sample-filename-2", fp2)
        self.assertEqual(
            {
                "Lausannevirus": {
                    "sample-filename-1": {
                        "proteins": {
                            "acc|GENBANK|I44.6|GENBANK|J77|VP1": {
                                "bestScore": 44.2,
                                "bluePlotFilename": "out/0.png",
                                "coverage": 0.63,
                                "readsFilename": "out/0.fasta",
                                "hspCount": 9,
                                "index": 0,
                                "medianScore": 41.3,
                                "outDir": "out",
                                "proteinLength": 12,
                                "proteinName": ("acc|GENBANK|I44.6|GENBANK|J77|VP1"),
                                "proteinURL": (
                                    "http://www.ncbi.nlm.nih.gov/" "nuccore/I44.6"
                                ),
                                "genomeURL": (
                                    "http://www.ncbi.nlm.nih.gov/nuccore/J77"
                                ),
                                "readCount": 9,
                                "readAndHspCountStr": "9",
                            },
                        },
                        "uniqueReadCount": None,
                    },
                    "sample-filename-2": {
                        "proteins": {
                            "acc|GENBANK|I44.7|GENBANK|J78|VP2": {
                                "bestScore": 48.1,
                                "bluePlotFilename": "out/0.png",
                                "coverage": 0.77,
                                "readsFilename": "out/0.fasta",
                                "hspCount": 6,
                                "index": 0,
                                "medianScore": 46.6,
                                "outDir": "out",
                                "proteinLength": 74,
                                "proteinName": ("acc|GENBANK|I44.7|GENBANK|J78|VP2"),
                                "proteinURL": (
                                    "http://www.ncbi.nlm.nih.gov/" "nuccore/I44.7"
                                ),
                                "genomeURL": (
                                    "http://www.ncbi.nlm.nih.gov/nuccore/J78"
                                ),
                                "readCount": 5,
                                "readAndHspCountStr": "5/6",
                            },
                        },
                        "uniqueReadCount": None,
                    },
                },
            },
            pg.pathogenNames,
        )

    def testOneLineInEachOfTwoFilesSamePathogenTitle(self):
        """
        If a protein grouper is given two files, each with one line from the
        same pathogen, its _title method must return the expected string.
        """
        fp1 = StringIO(
            "0.63 41.3 44.2 9 9 12 acc|GENBANK|I44.6|GENBANK|J77|VP1 "
            "[Lausannevirus]\n"
        )
        fp2 = StringIO(
            "0.77 46.6 48.1 5 6 74 acc|GENBANK|I44.7|GENBANK|J78|VP2 "
            "[Lausannevirus]\n"
        )
        pg = ProteinGrouper()
        pg.addFile("sample-filename-1", fp1)
        pg.addFile("sample-filename-2", fp2)
        self.assertEqual(
            "Overall, proteins from 1 pathogen were found in 2 samples.", pg._title()
        )

    def testOneLineInEachOfTwoFilesDifferentPathogens(self):
        """
        If a protein grouper is given two files in two different directories,
        each with one line from the different pathogens, its pathogenNames dict
        must be as expected.
        """
        fp1 = StringIO(
            "0.63 41.3 44.2 9 9 12 acc|GENBANK|I44.6|GENBANK|J77|VP1 "
            "[Lausannevirus]\n"
        )
        fp2 = StringIO(
            "0.77 46.6 48.1 5 6 74 acc|GENBANK|I44.7|GENBANK|J78|VP2 "
            "[Hepatitis B virus]\n"
        )
        pg = ProteinGrouper()
        pg.addFile("dir-1/sample-filename-1", fp1)
        pg.addFile("dir-2/sample-filename-2", fp2)
        self.assertEqual(
            {
                "Lausannevirus": {
                    "dir-1/sample-filename-1": {
                        "proteins": {
                            "acc|GENBANK|I44.6|GENBANK|J77|VP1": {
                                "bestScore": 44.2,
                                "bluePlotFilename": "dir-1/out/0.png",
                                "coverage": 0.63,
                                "readsFilename": "dir-1/out/0.fasta",
                                "hspCount": 9,
                                "index": 0,
                                "medianScore": 41.3,
                                "outDir": "dir-1/out",
                                "proteinLength": 12,
                                "proteinName": ("acc|GENBANK|I44.6|GENBANK|J77|VP1"),
                                "proteinURL": (
                                    "http://www.ncbi.nlm.nih.gov/" "nuccore/I44.6"
                                ),
                                "genomeURL": (
                                    "http://www.ncbi.nlm.nih.gov/nuccore/J77"
                                ),
                                "readCount": 9,
                                "readAndHspCountStr": "9",
                            },
                        },
                        "uniqueReadCount": None,
                    },
                },
                "Hepatitis B virus": {
                    "dir-2/sample-filename-2": {
                        "proteins": {
                            "acc|GENBANK|I44.7|GENBANK|J78|VP2": {
                                "bestScore": 48.1,
                                "bluePlotFilename": "dir-2/out/0.png",
                                "coverage": 0.77,
                                "readsFilename": "dir-2/out/0.fasta",
                                "hspCount": 6,
                                "index": 0,
                                "medianScore": 46.6,
                                "outDir": "dir-2/out",
                                "proteinLength": 74,
                                "proteinName": ("acc|GENBANK|I44.7|GENBANK|J78|VP2"),
                                "proteinURL": (
                                    "http://www.ncbi.nlm.nih.gov/" "nuccore/I44.7"
                                ),
                                "genomeURL": (
                                    "http://www.ncbi.nlm.nih.gov/nuccore/J78"
                                ),
                                "readCount": 5,
                                "readAndHspCountStr": "5/6",
                            },
                        },
                        "uniqueReadCount": None,
                    },
                },
            },
            pg.pathogenNames,
        )

    def testOneLineInEachOfTwoFilesDifferentPathogensTitle(self):
        """
        If a protein grouper is given two files, each with one line from
        different pathogens, its _title method must return the expected string.
        """
        fp1 = StringIO(
            "0.63 41.3 44.2 9 9 12 acc|GENBANK|I44.6|GENBANK|J77|VP1 "
            "[Lausannevirus]\n"
        )
        fp2 = StringIO(
            "0.77 46.6 48.1 5 6 74 acc|GENBANK|I44.7|GENBANK|J78|VP2 "
            "[Hepatitis B virus]\n"
        )
        pg = ProteinGrouper()
        pg.addFile("sample-filename-1", fp1)
        pg.addFile("sample-filename-2", fp2)
        self.assertEqual(
            "Overall, proteins from 2 pathogens were found in 2 samples.", pg._title()
        )

    def testNoFilesToStr(self):
        """
        If no files have been given to a protein grouper, its text string
        format must as expected.
        """
        pg = ProteinGrouper()
        self.assertEqual(
            "Summary of pathogens\n"
            "\n"
            "Overall, proteins from 0 pathogens were found in 0 samples.\n",
            pg.toStr(),
        )

    def testNoFilesToStrWithTitle(self):
        """
        If no files have been given to a protein grouper, its text string
        format (including a passed title) must as expected.
        """
        pg = ProteinGrouper()
        self.assertEqual(
            "The title...\n"
            "\n"
            "Overall, proteins from 0 pathogens were found in 0 samples.\n",
            pg.toStr(title="The title..."),
        )

    def testNoFilesToStrWithTitleAndPreamble(self):
        """
        If no files have been given to a protein grouper, its text string
        format (including a passed title and preamble) must as expected.
        """
        pg = ProteinGrouper()
        self.assertEqual(
            "The title...\n"
            "\n"
            "The preamble...\n"
            "\n"
            "Overall, proteins from 0 pathogens were found in 0 samples.\n",
            pg.toStr(title="The title...", preamble="The preamble..."),
        )

    def testOneLineInOneFileToStr(self):
        """
        If a protein grouper is given one file with one line, its toStr method
        must produce the expected result.
        """
        fp = StringIO(
            "0.77 46.6 48.1 5 6 74 acc|GENBANK|I44.6|GENBANK|J77|"
            "VP1 [Lausannevirus]\n"
        )
        pg = ProteinGrouper()
        pg.addFile("sample-filename", fp)
        self.assertEqual(
            "Summary of pathogens\n"
            "\n"
            "Overall, proteins from 1 pathogen were found in 1 sample.\n"
            "\n"
            "Lausannevirus (in 1 sample)\n"
            "  sample-filename (1 protein, 5 reads)\n"
            "    0.77\t46.60\t48.10\t        5/6\tacc|"
            "GENBANK|I44.6|GENBANK|J77|VP1\n",
            pg.toStr(),
        )

    def testPathogenNameRegex(self):
        """
        If a protein grouper is given one file with two lines from different
        pathogens, and one pathogen title does not match a passed title regex,
        the pathogenNames dict must be as expected.
        """
        fp = StringIO(
            "0.63 41.3 44.2 9 9 12 acc|GENBANK|I44.6|GENBANK|J77|VP1 "
            "[Lausannevirus]\n"
            "0.77 46.6 48.1 5 6 74 acc|GENBANK|I44.7|GENBANK|J78|VP2 "
            "[Hepatitis B virus]\n"
        )
        pg = ProteinGrouper(titleRegex="Lausannevirus")
        pg.addFile("sample-filename", fp)
        self.assertEqual(
            {
                "Lausannevirus": {
                    "sample-filename": {
                        "proteins": {
                            "acc|GENBANK|I44.6|GENBANK|J77|VP1": {
                                "bestScore": 44.2,
                                "bluePlotFilename": "out/0.png",
                                "coverage": 0.63,
                                "readsFilename": "out/0.fasta",
                                "hspCount": 9,
                                "index": 0,
                                "medianScore": 41.3,
                                "outDir": "out",
                                "proteinLength": 12,
                                "proteinName": ("acc|GENBANK|I44.6|GENBANK|J77|VP1"),
                                "proteinURL": (
                                    "http://www.ncbi.nlm.nih.gov/" "nuccore/I44.6"
                                ),
                                "genomeURL": (
                                    "http://www.ncbi.nlm.nih.gov/nuccore/J77"
                                ),
                                "readCount": 9,
                                "readAndHspCountStr": "9",
                            },
                        },
                        "uniqueReadCount": None,
                    },
                },
            },
            pg.pathogenNames,
        )

    def testPathogenNegativeNameRegex(self):
        """
        If a protein grouper is given one file with two lines from different
        pathogens, and one pathogen title does not match a passed negative
        title regex, the pathogenNames dict must be as expected.
        """
        fp = StringIO(
            "0.63 41.3 44.2 9 9 12 acc|GENBANK|I44.6|GENBANK|J77|VP1 "
            "[Lausannevirus]\n"
            "0.77 46.6 48.1 5 6 74 acc|GENBANK|I44.7|GENBANK|J78|VP2 "
            "[Hepatitis B virus]\n"
        )
        pg = ProteinGrouper(negativeTitleRegex="Hepatitis")
        pg.addFile("sample-filename", fp)
        self.assertEqual(
            {
                "Lausannevirus": {
                    "sample-filename": {
                        "proteins": {
                            "acc|GENBANK|I44.6|GENBANK|J77|VP1": {
                                "bestScore": 44.2,
                                "bluePlotFilename": "out/0.png",
                                "coverage": 0.63,
                                "readsFilename": "out/0.fasta",
                                "hspCount": 9,
                                "index": 0,
                                "medianScore": 41.3,
                                "outDir": "out",
                                "proteinLength": 12,
                                "proteinName": ("acc|GENBANK|I44.6|GENBANK|J77|VP1"),
                                "proteinURL": (
                                    "http://www.ncbi.nlm.nih.gov/" "nuccore/I44.6"
                                ),
                                "genomeURL": (
                                    "http://www.ncbi.nlm.nih.gov/nuccore/J77"
                                ),
                                "readCount": 9,
                                "readAndHspCountStr": "9",
                            },
                        },
                        "uniqueReadCount": None,
                    },
                },
            },
            pg.pathogenNames,
        )


class TestPathogenSampleFiles(TestCase):
    """
    Tests for the PathogenSampleFiles class.
    """

    def testUnknownFormat(self):
        """
        Passing an unknown format argument must result in a ValueError
        being raised.
        """
        pg = ProteinGrouper()
        error = r"^format_ must be either 'fasta' or 'fastq'\.$"
        assertRaisesRegex(
            self, ValueError, error, PathogenSampleFiles, pg, format_="unknown"
        )

    @skip("Some tests are broken and skipped under latest BioPython")
    def testOpenNotCalledOnRepeatedCall(self):
        """
        If a repeated call to pathogenSampleFiles.add is made with the same
        arguments, no file should be read because the original result value is
        cached.
        """

        class Open:
            def __init__(self, test, manager):
                self.test = test
                self.manager = manager
                self.count = 0

            def sideEffect(self, filename, *args, **kwargs):
                if self.count == 0:
                    self.test.assertEqual("out/0.fasta", filename)
                    self.count += 1
                    return StringIO(">id1\nACTG\n")
                elif self.count == 1:
                    self.test.assertEqual("out/pathogen-0-sample-0.fasta", filename)
                    self.count += 1
                    return self.manager
                else:
                    self.test.fail(
                        "We are only supposed to be called twice. "
                        "Filename: %r, Args: %r, Keyword args: %r."
                        % (filename, args, kwargs)
                    )

        fp = StringIO(
            "0.77 46.6 48.1 5 6 74 acc|GENBANK|I44.6|GENBANK|J77|"
            "VP1 [Lausannevirus]\n"
        )
        fastaIO = StringIO()

        @contextmanager
        def manager():
            yield fastaIO

        pg = ProteinGrouper()
        pg.addFile("filename-1", fp)
        pathogenSampleFiles = PathogenSampleFiles(pg)

        sideEffect = Open(self, manager()).sideEffect
        with patch.object(builtins, "open") as mockMethod:
            mockMethod.side_effect = sideEffect
            filename = pathogenSampleFiles.add("Lausannevirus", "filename-1")
            self.assertEqual("out/pathogen-0-sample-0.fasta", filename)
            self.assertEqual(">id1\nACTG\n", fastaIO.getvalue())

            # Repeated call. The side effect open will fail if open is
            # called at this point.
            filename = pathogenSampleFiles.add("Lausannevirus", "filename-1")
            self.assertEqual("out/pathogen-0-sample-0.fasta", filename)

    @skip("Some tests are broken and skipped under latest BioPython")
    def testIdenticalReadsRemoved(self):
        """
        If two proteins in the same pathogen are matched by the same read, the
        de-duplicated FASTA for the pathogen must have only one copy of the
        duplicated read.
        """

        class Open:
            def __init__(self, test, manager):
                self.test = test
                self.manager = manager
                self.expectedFilenames = {
                    "out/0.fasta",
                    "out/1.fasta",
                    "out/pathogen-0-sample-0.fasta",
                }

            def sideEffect(self, filename, *args, **kwargs):
                try:
                    self.expectedFilenames.remove(filename)
                except KeyError:
                    self.test.fail(
                        "Open called with unexpected filename: %r, Args: %r, "
                        "Keyword args: %r." % (filename, args, kwargs)
                    )
                else:
                    if filename == "out/0.fasta":
                        return StringIO(">id1\nACTG\n")
                    elif filename == "out/1.fasta":
                        return StringIO(">id1\nACTG\n>id2\nCAGT\n")
                    else:
                        return self.manager

        fp = StringIO(
            "0.63 41.3 44.2 9 9 12 acc|GENBANK|I44.6|GENBANK|J77|VP1 "
            "[Lausannevirus]\n"
            "0.77 46.6 48.1 5 6 74 acc|GENBANK|I44.7|GENBANK|J78|VP2 "
            "[Lausannevirus]\n"
        )
        fastaIO = StringIO()

        @contextmanager
        def manager():
            yield fastaIO

        pg = ProteinGrouper()
        pg.addFile("filename-1", fp)
        pathogenSampleFiles = PathogenSampleFiles(pg)

        opener = Open(self, manager())
        with patch.object(builtins, "open") as mockMethod:
            mockMethod.side_effect = opener.sideEffect
            filename = pathogenSampleFiles.add("Lausannevirus", "filename-1")

        self.assertEqual("out/pathogen-0-sample-0.fasta", filename)
        self.assertEqual(">id1\nACTG\n>id2\nCAGT\n", fastaIO.getvalue())
        # Make sure all expected filenames were seen by the mocked open.
        self.assertEqual(set(), opener.expectedFilenames)

    @skip("Some tests are broken and skipped under latest BioPython")
    def testPathogenIndex(self):
        """
        A pathogen index must be retrievable via the pathogenIndex function.
        """

        class Open:
            def __init__(self, test, manager):
                self.test = test
                self.manager = manager
                self.expectedFilenames = {
                    "out/0.fasta",
                    "out/1.fasta",
                    "out/pathogen-0-sample-0.fasta",
                }

            def sideEffect(self, filename, *args, **kwargs):
                try:
                    self.expectedFilenames.remove(filename)
                except KeyError:
                    self.test.fail(
                        "Open called with unexpected filename: %r, Args: %r, "
                        "Keyword args: %r." % (filename, args, kwargs)
                    )
                else:
                    if filename == "out/0.fasta":
                        return StringIO(">id1\nACTG\n")
                    elif filename == "out/1.fasta":
                        return StringIO(">id1\nACTG\n>id2\nCAGT\n")
                    else:
                        return self.manager

        fp = StringIO(
            "0.63 41.3 44.2 9 9 12 acc|GENBANK|I44.6|GENBANK|J77|VP1 "
            "[Lausannevirus]\n"
            "0.77 46.6 48.1 5 6 74 acc|GENBANK|I44.7|GENBANK|J78|VP2 "
            "[Lausannevirus]\n"
        )
        fastaIO = StringIO()

        @contextmanager
        def manager():
            yield fastaIO

        pg = ProteinGrouper()
        pg.addFile("filename-1", fp)
        pathogenSampleFiles = PathogenSampleFiles(pg)

        opener = Open(self, manager())
        with patch.object(builtins, "open") as mockMethod:
            mockMethod.side_effect = opener.sideEffect
            filename = pathogenSampleFiles.add("Lausannevirus", "filename-1")

        # Leaving this line in so we can explicitly see that the first
        # pathogen has index zero.
        self.assertEqual("out/pathogen-0-sample-0.fasta", filename)

        # Here's the pathogen index we're actually interested in testing.
        self.assertEqual(0, pathogenSampleFiles.pathogenIndex("Lausannevirus"))

        # Make sure all expected filenames were seen by the mocked open.
        self.assertEqual(set(), opener.expectedFilenames)

    @skip("Some tests are broken and skipped under latest BioPython")
    def testProteinsSavedCorrectly(self):
        """
        Information about proteins must be saved correctly in the
        ProteinGrouper for a given pathogen/sample combination.
        """

        class Open:
            def __init__(self, test, manager):
                self.test = test
                self.manager = manager
                self.expectedFilenames = {
                    "out/0.fasta",
                    "out/1.fasta",
                    "out/pathogen-0-sample-0.fasta",
                }

            def sideEffect(self, filename, *args, **kwargs):
                if filename in self.expectedFilenames:
                    if filename == "out/0.fasta":
                        return StringIO(">id1\nACTG\n")
                    elif filename == "out/1.fasta":
                        return StringIO(">id2\nAC\n>id3\nCAGTTTT\n")
                    else:
                        return self.manager
                else:
                    self.test.fail(
                        "Open called with unexpected filename: %r, Args: %r, "
                        "Keyword args: %r." % (filename, args, kwargs)
                    )

        fp = StringIO(
            "0.63 41.3 44.2 9 9 12 acc|GENBANK|I44.6|GENBANK|J77|VP1 "
            "[Lausannevirus]\n"
            "0.77 46.6 48.1 5 6 74 acc|GENBANK|I44.7|GENBANK|J78|VP2 "
            "[Lausannevirus]\n"
        )
        fastaIO = StringIO()

        @contextmanager
        def manager():
            yield fastaIO

        opener = Open(self, manager())
        with patch.object(builtins, "open") as mockMethod:
            mockMethod.side_effect = opener.sideEffect
            pg = ProteinGrouper()
            pg.addFile("filename-1", fp)
            pathogenSampleFiles = PathogenSampleFiles(pg)
            pathogenSampleFiles.add("Lausannevirus", "filename-1")

        self.assertEqual(
            {
                "proteins": {
                    "acc|GENBANK|I44.6|GENBANK|J77|VP1": {
                        "bestScore": 44.2,
                        "bluePlotFilename": "out/0.png",
                        "coverage": 0.63,
                        "readsFilename": "out/0.fasta",
                        "hspCount": 9,
                        "index": 0,
                        "medianScore": 41.3,
                        "outDir": "out",
                        "proteinLength": 12,
                        "proteinName": ("acc|GENBANK|I44.6|GENBANK|J77|VP1"),
                        "proteinURL": ("http://www.ncbi.nlm.nih.gov/" "nuccore/I44.6"),
                        "genomeURL": ("http://www.ncbi.nlm.nih.gov/nuccore/J77"),
                        "readCount": 9,
                        "readAndHspCountStr": "9",
                    },
                    "acc|GENBANK|I44.7|GENBANK|J78|VP2": {
                        "bestScore": 48.1,
                        "bluePlotFilename": "out/1.png",
                        "coverage": 0.77,
                        "readsFilename": "out/1.fasta",
                        "hspCount": 6,
                        "index": 1,
                        "medianScore": 46.6,
                        "outDir": "out",
                        "proteinLength": 74,
                        "proteinName": ("acc|GENBANK|I44.7|GENBANK|J78|VP2"),
                        "proteinURL": ("http://www.ncbi.nlm.nih.gov/" "nuccore/I44.7"),
                        "genomeURL": ("http://www.ncbi.nlm.nih.gov/nuccore/J78"),
                        "readCount": 5,
                        "readAndHspCountStr": "5/6",
                    },
                },
                "uniqueReadCount": 3,
            },
            pg.pathogenNames["Lausannevirus"]["filename-1"],
        )

    @skip("Some tests are broken and skipped under latest BioPython")
    def testReadLengthsAdded(self, unlinkMock):
        """
        If saveReadLengths is True for a ProteinGrouper, read lengths must be
        saved for each protein.
        """

        class Open:
            def __init__(self, test, manager):
                self.test = test
                self.manager = manager
                self.expectedFilenames = {
                    "out/0.fasta",
                    "out/1.fasta",
                    "out/pathogen-0-sample-0.fasta",
                }

            def sideEffect(self, filename, *args, **kwargs):
                if filename in self.expectedFilenames:
                    if filename == "out/0.fasta":
                        return StringIO(">id1\nACTG\n")
                    elif filename == "out/1.fasta":
                        return StringIO(">id2\nAC\n>id3\nCAGTTTT\n")
                    else:
                        return self.manager
                else:
                    self.test.fail(
                        "Open called with unexpected filename: %r, Args: %r, "
                        "Keyword args: %r." % (filename, args, kwargs)
                    )

        fp = StringIO(
            "0.63 41.3 44.2 9 9 12 acc|GENBANK|I44.6|GENBANK|J77|VP1 "
            "[Lausannevirus]\n"
            "0.77 46.6 48.1 5 6 74 acc|GENBANK|I44.7|GENBANK|J78|VP2 "
            "[Lausannevirus]\n"
        )
        fastaIO = StringIO()

        @contextmanager
        def manager():
            yield fastaIO

        opener = Open(self, manager())
        with patch.object(builtins, "open") as mockMethod:
            mockMethod.side_effect = opener.sideEffect
            pg = ProteinGrouper(saveReadLengths=True)
            pg.addFile("filename-1", fp)
            pathogenSampleFiles = PathogenSampleFiles(pg)
            pathogenSampleFiles.add("Lausannevirus", "filename-1")

        # Read lengths must be saved correctly.
        proteins = pg.pathogenNames["Lausannevirus"]["filename-1"]["proteins"]
        self.assertEqual(
            (4,), proteins["acc|GENBANK|I44.6|GENBANK|J77|VP1"]["readLengths"]
        )
        self.assertEqual(
            (2, 7), proteins["acc|GENBANK|I44.7|GENBANK|J78|VP2"]["readLengths"]
        )

    def testWriteSampleIndex(self):
        """
        The writeSampleIndex function must write a file with the expected
        content.
        """
        pathogenSampleFiles = PathogenSampleFiles(None)
        pathogenSampleFiles._samples = {
            "NEO11": 500,
            "NEO33": 24,
            "NEO66": 333,
        }

        fp = StringIO()
        pathogenSampleFiles.writeSampleIndex(fp)
        self.assertEqual("24 NEO33\n333 NEO66\n500 NEO11\n", fp.getvalue())

    def testWritePathogenIndex(self):
        """
        The writePatogenIndex function must write a file with the expected
        content.
        """
        pathogenSampleFiles = PathogenSampleFiles(None)
        pathogenSampleFiles._pathogens = {
            "virus b": 4,
            "virus a": 3,
            "virus c": 9,
        }

        fp = StringIO()
        pathogenSampleFiles.writePathogenIndex(fp)
        self.assertEqual("3 virus a\n4 virus b\n9 virus c\n", fp.getvalue())



from typing import Generator, Union, Iterable, Mapping, Optional
from itertools import zip_longest

from dark.aaVars import (
    ABBREV3,
    ABBREV3_TO_ABBREV1,
    CODONS,
    NAMES,
    NAMES_TO_ABBREV1,
    PROPERTIES,
    PROPERTY_CLUSTERS,
    PROPERTY_DETAILS,
)

from dark.reads import Read
from dark.utils import countPrint


class AminoAcid:
    """
    Hold information about an amino acid.

    @param name: The full C{str} name of the amino acid.
    @param abbrev3: The 3-letter C{str} abbreviation of the amino acid,
        e.g., 'Arg'.
    @param abbrev1: The 1-letter C{str} abbreviation of the amino acid,
        e.g., 'A'.
    @param codons: A C{list} of 3-letter codons for the amino acid.
    @param properties: An C{int} logical-AND of the various properties
        (see PROPERTIES, above) of this amino acid.
    @param propertyDetails: A C{dict} containing property names and values
        for this amino acid. E.g.:
        {
            'aliphaticity': -0.157024793388,
            'aromaticity': -0.0642673521851,
            'composition': -0.527272727273,
            'hydrogenation': -0.401797175866,
            'hydropathy': -1.0,
            'hydroxythiolation': -0.51486325802,
            'iep': 1.0,
            'polar requirement': 0.0487804878049,
            'polarity': 0.382716049383,
            'volume': 0.449101796407,
        }
    @param propertyClusters: A C{dict} containing the property names and
        clusters for this amino acid. E.g.:
        {
            'aliphaticity': 1,
            'aromaticity': 2,
            'composition': 1,
            'hydrogenation': 1,
            'hydropathy': 2,
            'hydroxythiolation': 4,
            'iep': 2,
            'polar requirement': 1,
            'polarity': 1,
            'volume': 4,
        }
    """

    def __init__(
        self,
        name: str,
        abbrev3: str,
        abbrev1: str,
        codons: tuple[str, ...],
        properties: int,
        propertyDetails: dict[str, float],
        propertyClusters: dict[str, int],
    ):
        self.name = name
        self.abbrev3 = abbrev3
        self.abbrev1 = abbrev1
        self.codons = codons
        self.properties = properties
        self.propertyDetails = propertyDetails
        self.propertyClusters = propertyClusters


def find(s: str) -> Generator[AminoAcid, None, None]:
    """
    Find an amino acid whose name or abbreviation is s.

    @param s: A C{str} amino acid specifier. This may be a full name,
        a 3-letter abbreviation or a 1-letter abbreviation. Case is ignored.
    return: A generator that yields matching L{AminoAcid} instances.
    """

    abbrev1 = None
    origS = s

    if " " in s:
        # Convert first word to title case, others to lower.
        first, rest = s.split(" ", 1)
        s = first.title() + " " + rest.lower()
    else:
        s = s.title()

    if s in NAMES:
        abbrev1 = s
    elif s in ABBREV3_TO_ABBREV1:
        abbrev1 = ABBREV3_TO_ABBREV1[s]
    elif s in NAMES_TO_ABBREV1:
        abbrev1 = NAMES_TO_ABBREV1[s]
    else:
        # Look for a 3-letter codon.
        def findCodon(target):
            for abbrev1, codons in CODONS.items():
                for codon in codons:
                    if codon == target:
                        return abbrev1

        abbrev1 = findCodon(origS.upper())

    if abbrev1:
        abbrev1s = [abbrev1]
    else:
        # Try partial matching on names.
        abbrev1s = []
        sLower = s.lower()
        for abbrev1, name in NAMES.items():
            if name.lower().find(sLower) > -1:
                abbrev1s.append(abbrev1)

    for abbrev1 in abbrev1s:
        yield AminoAcid(
            NAMES[abbrev1],
            ABBREV3[abbrev1],
            abbrev1,
            CODONS[abbrev1],
            PROPERTIES[abbrev1],
            PROPERTY_DETAILS[abbrev1],
            PROPERTY_CLUSTERS[abbrev1],
        )


def _propertiesOrClustersForSequence(
    sequence: Read,
    propertyNames: Iterable[str],
    propertyValues: Mapping[str, Mapping[str, Union[int, float]]],
    missingAAValue: Union[int, float],
) -> dict[str, list[Union[int, float]]]:
    """
    Extract amino acid property values or cluster numbers for a sequence.

    @param sequence: An C{AARead} (or a subclass) instance.
    @param propertyNames: An iterable of C{str} property names (each of which
        must be a key of a key in the C{propertyValues} C{dict}).
    @param propertyValues: A C{dict} in the form of C{PROPERTY_DETAILS} or
        C{PROPERTY_CLUSTERS} (see above).
    @param missingAAValue: A C{float} value to use for properties when an AA
        (e.g., 'X') is not known.
    @raise ValueError: If an unknown property is given in C{propertyNames}.
    @return: A C{dict} keyed by (lowercase) property name, with values that are
        C{list}s of the corresponding C{float} property value in C{propertyValues} in
        order of sequence position.
    """
    propertyNames = sorted(map(str.lower, set(propertyNames)))

    # Make sure all mentioned property names exist for at least one AA.
    knownProperties: set[str] = set()
    for names in propertyValues.values():
        knownProperties.update(names)
    unknown = set(propertyNames) - knownProperties
    if unknown:
        raise ValueError(
            "Unknown propert%s: %s."
            % ("y" if len(unknown) == 1 else "ies", ", ".join(unknown))
        )

    aas = sequence.sequence.upper()
    result: dict[str, list[float]] = {}

    for propertyName in propertyNames:
        result[propertyName] = []
        append = result[propertyName].append
        for aa in aas:
            try:
                properties = propertyValues[aa]
            except KeyError:
                # No such AA.
                append(missingAAValue)
            else:
                append(properties[propertyName])

    return result


def propertiesForSequence(
    sequence: Read, propertyNames: Iterable[str], missingAAValue: float = -1.1
):
    """
    Extract amino acid property values for a sequence.

    @param sequence: An C{AARead} (or a subclass) instance.
    @param propertyNames: An iterable of C{str} property names (each of which
        must be a key of a key in the C{dark.aa.PROPERTY_DETAILS} C{dict}).
    @param missingAAValue: A C{float} value to use for properties when an AA
        (e.g., 'X') is not known.
    @raise ValueError: If an unknown property is given in C{propertyNames}.
    @return: A C{dict} keyed by (lowercase) property name, with values that are
        C{list}s of the corresponding property value according to sequence
        position.
    """
    return _propertiesOrClustersForSequence(
        sequence, propertyNames, PROPERTY_DETAILS, missingAAValue
    )


def clustersForSequence(
    sequence: Read, propertyNames: Iterable[str], missingAAValue: int = 0
):
    """
    Extract amino acid property cluster numbers for a sequence.

    @param sequence: An C{AARead} (or a subclass) instance.
    @param propertyNames: An iterable of C{str} property names (each of which
        must be a key of a key in the C{dark.aa.PROPERTY_CLUSTERS} C{dict}).
    @param missingAAValue: An C{int} value to use for properties when an AA
        (e.g., 'X') is not known.
    @raise ValueError: If an unknown property is given in C{propertyNames}.
    @return: A C{dict} keyed by (lowercase) property name, with values that are
        C{list}s of the corresponding C{int} property cluster number according to
        sequence position.
    """
    return _propertiesOrClustersForSequence(
        sequence, propertyNames, PROPERTY_CLUSTERS, missingAAValue
    )


# It would be nice, in some universe, to do the following. But we access the
# match dict using a variable key and mypy cannot check that, so this doesn't work.
# Leaving it here in case it can be used for stricter checking in the future.
#
# _OverallMatch = TypedDict(
#     "_OverallMatch",
#     {
#         "matchCount": int,
#         "gapMismatchCount": int,
#         "gapGapMismatchCount": int,
#         "nonGapMismatchCount": int,
#     },
# )
#
# _Read1Match = TypedDict(
#     "_Read1Match",
#     {
#         "extraCount": int,
#         "gapOffsets": list[int],
#     },
# )
#
# _Read2Match = TypedDict(
#     "_Read2Match",
#     {
#         "extraCount": int,
#         "gapOffsets": list[int],
#     },
# )
#
# Match = TypedDict(
#     "Match",
#     {
#         "match": _OverallMatch,
#         "read1": _Read1Match,
#         "read2": _Read2Match,
#     },
# )


# Be more relaxed.
_Match = Mapping[str, Mapping]


def matchToString(
    aaMatch: _Match,
    read1: Read,
    read2: Read,
    indent: str = "",
    offsets: Optional[set[int]] = None,
) -> str:
    """
    Format amino acid sequence match as a string.

    @param aaMatch: A C{dict} returned by C{compareAaReads}.
    @param read1: A C{Read} instance or an instance of one of its subclasses.
    @param read2: A C{Read} instance or an instance of one of its subclasses.
    @param indent: A C{str} to indent all returned lines with.
    @param offsets: If not C{None}, a C{set} of offsets of interest that were
        only considered when making C{match}.
    @return: A C{str} describing the match.
    """
    match = aaMatch["match"]
    matchCount = match["matchCount"]
    gapMismatchCount = match["gapMismatchCount"]
    gapGapMismatchCount = match["gapGapMismatchCount"]
    nonGapMismatchCount = match["nonGapMismatchCount"]

    if offsets:
        len1 = len2 = len(offsets)
    else:
        len1, len2 = map(len, (read1, read2))

    result: list[str] = []
    append = result.append

    append("%sMatches:" % indent)
    append(countPrint("%s  Overall" % indent, matchCount, len1, len2))
    append(
        countPrint(
            "%s  Not involving gaps (i.e., identities)" % indent,
            matchCount,
            matchCount + nonGapMismatchCount,
        )
    )

    append("%sMismatches:" % indent)
    mismatchCount = gapMismatchCount + gapGapMismatchCount + nonGapMismatchCount
    append(countPrint("%s  Overall" % indent, mismatchCount, len1, len2))
    append(
        countPrint(
            "%s  Not involving gaps (i.e., conflicts)" % indent,
            nonGapMismatchCount,
            len1,
            len2,
        )
    )
    append(
        countPrint(
            "%s  Involving a gap in one sequence" % indent, gapMismatchCount, len1, len2
        )
    )
    append(
        countPrint(
            "%s  Involving a gap in both sequences" % indent,
            gapGapMismatchCount,
            len1,
            len2,
        )
    )

    for read, key in zip((read1, read2), ("read1", "read2")):
        append("%sId: %s" % (indent, read.id))
        length = len(read)
        append("%s  Length: %d" % (indent, length))
        gapCount = len(aaMatch[key]["gapOffsets"])
        append(countPrint("%s  Gaps" % indent, gapCount, length))
        if gapCount:
            append(
                "%s  Gap locations (1-based): %s"
                % (
                    indent,
                    ", ".join(
                        map(
                            lambda offset: str(offset + 1),
                            sorted(aaMatch[key]["gapOffsets"]),
                        )
                    ),
                )
            )
        extraCount = aaMatch[key]["extraCount"]
        if extraCount:
            append(
                countPrint("%s  Extra amino acids at end" % indent, extraCount, length)
            )

    return "\n".join(result)


def compareAaReads(
    read1: Read, read2: Read, gapChars: str = "-", offsets: Optional[set[int]] = None
) -> _Match:
    """
    Compare two amino acid sequences.

    @param read1: A C{Read} instance or an instance of one of its subclasses.
    @param read2: A C{Read} instance or an instance of one of its subclasses.
    @param gapChars: An object supporting __contains__ with characters that
        should be considered to be gaps.
    @param offsets: If not C{None}, a C{set} of offsets of interest. Offsets
        not in the set will not be considered.
    @return: A C{dict} with information about the match and the individual
        sequences (see below).
    """
    matchCount = 0
    gapMismatchCount = nonGapMismatchCount = gapGapMismatchCount = 0
    read1ExtraCount = read2ExtraCount = 0
    read1GapOffsets = []
    read2GapOffsets = []

    for offset, (a, b) in enumerate(
        zip_longest(read1.sequence.upper(), read2.sequence.upper())
    ):
        # Use 'is not None' in the following to allow an empty offsets set
        # to be passed.
        if offsets is not None and offset not in offsets:
            continue
        if a is None:
            # b has an extra character at its end (it cannot be None).
            assert b is not None
            read2ExtraCount += 1
            if b in gapChars:
                read2GapOffsets.append(offset)
        elif b is None:
            # a has an extra character at its end.
            read1ExtraCount += 1
            if a in gapChars:
                read1GapOffsets.append(offset)
        else:
            # We have a character from both sequences (they could still be
            # gap characters).
            if a in gapChars:
                read1GapOffsets.append(offset)
                if b in gapChars:
                    # Both are gaps. This can happen (though hopefully not
                    # if the sequences were pairwise aligned).
                    gapGapMismatchCount += 1
                    read2GapOffsets.append(offset)
                else:
                    # a is a gap, b is not.
                    gapMismatchCount += 1
            else:
                if b in gapChars:
                    # b is a gap, a is not.
                    gapMismatchCount += 1
                    read2GapOffsets.append(offset)
                else:
                    # Neither is a gap character.
                    if a == b:
                        matchCount += 1
                    else:
                        nonGapMismatchCount += 1

    return {
        "match": {
            "matchCount": matchCount,
            "gapMismatchCount": gapMismatchCount,
            "gapGapMismatchCount": gapGapMismatchCount,
            "nonGapMismatchCount": nonGapMismatchCount,
        },
        "read1": {
            "extraCount": read1ExtraCount,
            "gapOffsets": read1GapOffsets,
        },
        "read2": {
            "extraCount": read2ExtraCount,
            "gapOffsets": read2GapOffsets,
        },
    }



#!/usr/bin/env python

"""
Read a file of our JSON BLAST output (which has one JSON object per
line) from stdin and pretty print it to stdout.
"""

from json import dumps, loads
import sys

for line in sys.stdin:
    s = dumps(loads(line[:-1]), sort_keys=True, indent=2)
    print("\n".join([field.rstrip() for field in s.splitlines()]))



#!/usr/bin/env python

import sys
import argparse

from dark.reads import addFASTACommandLineOptions, parseFASTACommandLineOptions, DNARead


FORWARD, REVERSE, COMPLEMENT, REVERSE_COMPLEMENT = 0, 1, 2, 3

SUFFIX = {FORWARD: "", REVERSE: " (r)", COMPLEMENT: " (c)", REVERSE_COMPLEMENT: " (rc)"}


parser = argparse.ArgumentParser(
    description=(
        "Given FASTA on stdin and a sequence (or genome region) "
        "to find, write information about matches (and, "
        "optionally, reverse complement matches), to standard "
        "ouput."
    )
)

parser.add_argument(
    "--reverseComplement",
    "--rc",
    action="store_true",
    help="Also search for the reverse complement. Reverse complement matches "
    f"are indicated with an{SUFFIX[REVERSE_COMPLEMENT]} suffix.",
)

parser.add_argument(
    "--reverse",
    action="store_true",
    help=(
        f"Also search for the reversed sequence. Reverse matches are "
        f"indicated with an{SUFFIX[REVERSE]} suffix."
    ),
)

parser.add_argument(
    "--complement",
    action="store_true",
    help=(
        f"Also search for the complemented sequence. Complement matches "
        f"are indicated with a{SUFFIX[COMPLEMENT]} suffix."
    ),
)

parser.add_argument(
    "--caseSensitive",
    action="store_true",
    help="Do not ignore upper/lower case when matching.",
)

parser.add_argument(
    "--printSequences",
    action="store_true",
    help=(
        f"Print the sequences searched for. Reversed, complemented, and "
        f"reverse-complemented sequences are indicated by "
        f"{SUFFIX[REVERSE]},{SUFFIX[COMPLEMENT]}, "
        f"and{SUFFIX[REVERSE_COMPLEMENT]}, respectively."
    ),
)

parser.add_argument(
    "--includeShortGenomes",
    action="store_true",
    help=(
        "Try to match region sequences that are shorter than "
        "the length of the given region (assuming they are "
        "non-empty) due to the input sequence being too short."
    ),
)

parser.add_argument(
    "--noShortGenomeWarning",
    dest="reportShortGenomes",
    action="store_false",
    help=(
        "Do not report (to standard error) when sequences are shorter "
        "than the length of the given region due to being too short."
    ),
)

parser.add_argument(
    "--printTrivialMatch",
    action="store_true",
    help=(
        "If --region is used, there is a guaranteed match at that "
        "location. Do not include that match in the printed matches "
        "unless this option is given."
    ),
)

parser.add_argument(
    "--sep", default="\t", help="The field separator string (default=TAB)."
)

parser.add_argument(
    "--zeroBased", action="store_true", help="Print zero-based matched offsets."
)

group = parser.add_mutually_exclusive_group(required=True)

group.add_argument(
    "--region",
    help=(
        "A region of the form A-B, where A and B are numeric sites, with "
        "A <= B. A and B will be treated as 1-based offsets unless "
        "--zeroBased is used."
    ),
)

group.add_argument("--sequence", help="The sequence to look for.")

addFASTACommandLineOptions(parser)
args = parser.parse_args()

if args.region:
    lo, hi = map(int, args.region.split("-"))

    if args.zeroBased:
        if not (0 <= lo <= hi):
            print(
                f"Invalid range {args.region!r}. Give two non-negative "
                f"numbers in non-decreasing order.",
                file=sys.stderr,
            )
            sys.exit(1)
    else:
        if not (0 < lo <= hi):
            extra = (
                " Or use --zeroBased if you really want to specify " "a zero offset."
                if lo == 0
                else ""
            )

            print(
                f"Invalid range {args.region!r}. Give two positive "
                f"numbers in non-decreasing order.{extra}",
                file=sys.stderr,
            )
            sys.exit(1)

        # Use zero-based indices.
        lo -= 1
else:
    target = args.sequence

reads = parseFASTACommandLineOptions(args)

for read in reads:
    if args.region:
        target = read.sequence[lo:hi]
        if (len(target) < hi - lo and not args.includeShortGenomes) or len(target) == 0:
            if args.reportShortGenomes:
                print(
                    f"The {read.id!r} sequence is too short ({len(read)}) "
                    f"for a full match of a region of length {hi - lo} "
                    f"starting at site {lo + 1} in the genome.",
                    file=sys.stderr,
                )
            continue

    # Always look for the given target.
    targets = [target]
    orientations = [FORWARD]

    # And look for target variants according to Boolean command-line options.
    if args.reverse:
        targets.append(target[::-1])
        orientations.append(REVERSE)

    if args.complement:
        targets.append(DNARead("id", target).reverseComplement().sequence[::-1])
        orientations.append(COMPLEMENT)

    if args.reverseComplement:
        targets.append(DNARead("id", target).reverseComplement().sequence)
        orientations.append(REVERSE_COMPLEMENT)

    sequence = read.sequence

    if not args.caseSensitive:
        # list is needed in the following as we may use 'targets'
        # twice.
        targets = list(map(str.upper, targets))

    matches = set()

    for thisTarget, orientation in zip(targets, orientations):
        start = 0
        while True:
            index = sequence.find(thisTarget, start)
            if index == -1:
                break
            else:
                if args.region is None or args.printTrivialMatch or index != lo:
                    assert (index, orientation) not in matches
                    matches.add((index, orientation))
                start = index + 1

    if matches:
        adjust = 0 if args.zeroBased else 1
        fields = [
            read.id,
            "%d match%s" % (len(matches), "" if len(matches) == 1 else "es"),
            ", ".join(
                str(index + adjust) + SUFFIX[orientation]
                for index, orientation in sorted(matches)
            ),
        ]

        if args.printSequences:
            for thisTarget, orientation in zip(targets, orientations):
                fields.append(("sequence%s = " % SUFFIX[orientation]) + thisTarget)

        print(args.sep.join(fields))



#!/usr/bin/env python

import sys

from dark.aa import find
from dark.codonDistance import codonInformation
from dark.aa import CODONS


def findOrDie(s):
    """
    Look up an amino acid.

    @param s: A C{str} amino acid specifier. This may be a full name,
        a 3-letter abbreviation or a 1-letter abbreviation. Case is ignored.
    @return: An C{AminoAcid} instance, if one can be found. Else exit.
    """
    aa = find(s)
    if aa:
        return aa
    else:
        print("Unknown amino acid or codon: %s" % s, file=sys.stderr)
        print("Valid arguments are: %s." % ", ".join(sorted(CODONS)), file=sys.stderr)
        sys.exit(1)


def codonKey(codonInfo):
    return codonInfo[0], codonInfo[1]


def transitionsKey(codonInfo):
    return codonInfo[2]


if len(sys.argv) != 3:
    print("Usage: %s amino_acid amino_acid" % sys.argv[0], file=sys.stderr)
    sys.exit(1)
else:
    (aa1,) = findOrDie(sys.argv[1])
    (aa2,) = findOrDie(sys.argv[2])

    distances = codonInformation(aa1.codons, aa2.codons, countTransitions=True)
    print("Codons:")
    print("%s: %s" % (aa1.name, ", ".join(sorted(aa1.codons))))
    print("%s: %s" % (aa2.name, ", ".join(sorted(aa2.codons))))

    # Output is sorted first by distance, then by number of transitions
    # (highest to lowest) then by codon DNA. The idea being to present the
    # possible codon changes to get from one amino acid to another in the
    # order that requires the least change to the most.

    print("Distances:")
    for distance in sorted(distances):
        for codon1, codon2, transitions in sorted(
            sorted(distances[distance], key=codonKey), key=transitionsKey, reverse=True
        ):
            print(
                f"{distance}: {aa1.name}: {codon1} -> "
                f"{aa2.name}: {codon2} (transitions: {transitions}, "
                f"transversions {distance - transitions})"
            )



from os.path import join

from dark.fastq import FastqReads
from dark.html import NCBISequenceLinkURL


class AlignmentPanelHTMLWriter:
    """
    Produces HTML details of a rectangular panel of graphs that each
    contain an alignment graph against a given sequence. This is
    supplementary output info for the AlignmentPanel class in graphics.py.

    @param outputDir: The C{str} directory to write files into.
    @param titlesAlignments: A L{dark.titles.TitlesAlignments} instance.
    """

    def __init__(self, outputDir, titlesAlignments):
        self._outputDir = outputDir
        self._titlesAlignments = titlesAlignments
        self._images = []

    def addImage(self, imageBasename, accession, title, graphInfo):
        self._images.append(
            {
                "accession": accession,
                "graphInfo": graphInfo,
                "imageBasename": imageBasename,
                "title": title,
            }
        )

    def close(self):
        with open(join(self._outputDir, "index.html"), "w") as fp:
            self._writeHeader(fp)
            self._writeBody(fp)
            self._writeFooter(fp)
        with open(join(self._outputDir, "style.css"), "w") as fp:
            self._writeCSS(fp)

    def _writeHeader(self, fp):
        fp.write(
            """\
<html>
  <head>
    <title>Read alignments for %d matched subjects</title>
    <link rel="stylesheet" type="text/css" href="style.css">
  </head>
  <body>
    <div id="content">
        """
            % len(self._images)
        )

    def _writeBody(self, fp):
        fp.write(
            "<h1>Read alignments for %d matched subjects</h1>\n" % len(self._images)
        )

        # Write out an alignment panel as a table.
        cols = 6
        fp.write("<table><tbody>\n")

        for i, image in enumerate(self._images):
            title = image["title"]
            accession = image["accession"]
            if i % cols == 0:
                fp.write("<tr>\n")

            fp.write(
                '<td><a id="small_%s"></a><a href="#big_%s"><img src="%s" '
                'class="thumbnail"/></a></td>\n'
                % (accession, accession, image["imageBasename"])
            )

            if i % cols == cols - 1:
                fp.write("</tr>")

        # Add empty cells to the final table row, and close the row, if
        # necessary.
        if i % cols < cols - 1:
            while i % cols < cols - 1:
                fp.write("<td>&nbsp;</td>\n")
                i += 1
            fp.write("</tr>\n")

        fp.write("</tbody></table>\n")

        # Write out the full images with additional detail.
        for i, image in enumerate(self._images):
            title = image["title"]
            accession = image["accession"]
            titleAlignments = self._titlesAlignments[title]
            graphInfo = image["graphInfo"]
            readFormat = self._writeReads(image)
            fp.write(
                """
      <a id="big_%s"></a>
      <h3>%d: %s</h3>
      <p>
            Length: %d.
            Read count: %d.
            HSP count: %d.
            <a href="%s.%s">%s</a>.
            <a href="#small_%s">Top panel.</a>
"""
                % (
                    accession,
                    i,
                    title,
                    titleAlignments.subjectLength,
                    titleAlignments.readCount(),
                    titleAlignments.hspCount(),
                    accession,
                    readFormat,
                    readFormat,
                    accession,
                )
            )

            url = NCBISequenceLinkURL(title)
            if url:
                fp.write('<a href="%s" target="_blank">NCBI</a>.' % url)

            # Write out feature information.
            if graphInfo["features"] is None:
                # Feature lookup was False (or we were offline).
                pass
            elif len(graphInfo["features"]) == 0:
                fp.write("There were no features.")
            else:
                fp.write('<a href="%s">Features</a>' % self._writeFeatures(i, image))

            # Write out the titles that this title invalidated due to its
            # read set.
            readSetFilter = self._titlesAlignments.readSetFilter
            if readSetFilter:
                invalidated = readSetFilter.invalidates(title)
                if invalidated:
                    nInvalidated = len(invalidated)
                    fp.write(
                        "<br/>This title invalidated %d other%s due to "
                        "its read set:<ul>"
                        % (nInvalidated, "" if nInvalidated == 1 else "s")
                    )
                    for title in invalidated:
                        fp.write("<li>%s</li>" % title)
                    fp.write("</ul>")

            fp.write('</p><img src="%s" class="full-size"/>' % image["imageBasename"])

    def _writeFooter(self, fp):
        fp.write(
            """\
    </div>
  </body>
</html>
"""
        )

    def _writeCSS(self, fp):
        fp.write(
            """\
#content {
  width: 95%;
  margin: auto;
}
img.thumbnail {
  height: 300px;
}
img.full-size {
  height: 900px;
}
"""
        )

    def _writeReads(self, image):
        """
        Write a FASTA or FASTQ file containing the set of reads that hit a
        sequence.

        @param image: A member of self._images.
        @return: A C{str}, either 'fasta' or 'fastq' indicating the format
            of the reads in C{self._titlesAlignments}.
        """
        if isinstance(self._titlesAlignments.readsAlignments.reads, FastqReads):
            format_ = "fastq"
        else:
            format_ = "fasta"
        filename = join(self._outputDir, "%s.%s" % (image["accession"], format_))
        titleAlignments = self._titlesAlignments[image["title"]]
        with open(filename, "w") as fp:
            for titleAlignment in titleAlignments:
                fp.write(titleAlignment.read.toString(format_))
        return format_

    def _writeFeatures(self, image):
        """
        Write a text file containing the features as a table.

        @param image: A member of self._images.
        @return: The C{str} features file name - just the base name, not
            including the path to the file.
        """
        basename = image["accession"] + "-features.txt"
        filename = join(self._outputDir, basename)
        featureList = image["graphInfo"]["features"]
        # Note that the following (deliberately) creates an empty features
        # file if there were no features.
        with open(filename, "w") as fp:
            for feature in featureList:
                fp.write("%s\n\n" % feature.feature)
        return basename



#!/usr/bin/env python

import sys
from Bio import SeqIO
from Bio.Seq import Seq
from dark.sequence import findPrimerBidiLimits


def trimPrimers(primer, verbose):
    """
    @param primer: A BioPython C{Bio.Seq} primer sequence.
    @param verbose: A C{bool}, if C{True} output additional information about
        how often and where primers were found.
    """
    reads = []
    absentCount = forwardCount = reverseCount = count = 0
    for seqRecord in SeqIO.parse(sys.stdin, "fasta"):
        count += 1
        start, end = findPrimerBidiLimits(primer, seqRecord.seq)
        if start == 0:
            if end == len(seqRecord):
                absentCount += 1
            else:
                reverseCount += 1
        else:
            forwardCount += 1
            if end != len(seqRecord):
                reverseCount += 1
        reads.append(seqRecord[start:end])

    if verbose:
        print(
            ("Read %d sequences. Found forward: %d, " "Found reversed: %d, Absent: %d")
            % (count, forwardCount, reverseCount, absentCount),
            file=sys.stderr,
        )

    SeqIO.write(reads, sys.stdout, "fasta")


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(
        description=(
            "Given FASTA on stdin, look for a primer sequence "
            "and write trimmed FASTA (after the primer) to stdout."
        )
    )

    parser.add_argument("primer", help="the primer sequence")
    parser.add_argument(
        "--verbose",
        type=bool,
        default=False,
        help="If True, print information on found primers",
    )

    args = parser.parse_args()

    trimPrimers(Seq(args.primer.upper()), args.verbose)
